#!/usr/bin/perl


#-----------------------------------------------------------------------------------------
#
# Copyright (c) 2000 Conexant Systems
$HelpText = <<__EOH__;
#
#NAME
#	hsf_linstall.pl - initializes modem configuration
#
#SYNOPSIS
#	hsf_linstall.pl [-c <country-name|country-code>] [-help] [-s] [-t] 
#	                [-u|usage] [-v|version] [<release-dir>]
#
#DESCRIPTION
#	The program searches for SoftK56 PCI Modem installed in the system
#	(according to the /proc/bus/pci/devices or /proc/pci system files) and prepares for 
#	the modem driver the initialization data according to the corresponding INF 
#	configuration file.
#	
#	In addition the program gives the option to modify the CURRENT 
#	country in the INF configuration file, if <country-name> is
#	supported by the INF File, and to see the list of all supported countries.
#
#	The INF configuration file is taken from the <release-dir> in case it is specified in the
#	command line. Otherwise <release-dir> is assumed to be the current directory.
#
#	OPTIONS
#		Any option should be preceded by the '-' character.
#		Some options have two equivalent names, one is a single letter
#		and the other is a long name.
#		
#		Long name options can be abbreviated to any unique prefix of their name 
#		(for example : 'hsf_linstall.pl -ver' instead of 'hsf_linstall.pl -version' ).
#		Multiple single letter options can NOT be combined into a single command 
#		line word.
#
#		-c      modify the CURRENT country in the INF file to <country-name> and exit
#
#		-help   display this help and exit
#
#		-s      list supported countries by the INF modem configuration file, mark the
#		        CURRENT country and exit
#
#		-t      list all supported countries by the Initialization program and exit.
#		        It is possible that supported country by the program would not be
#		        supported by the INF configuration file.
#					
#		-u, -usage
#		        output usage information and exit  
#				
#		-v, -version 
#		        output version information and exit  
#
#EXAMPLES
#	Simple Initialization
#		> hsf_linstall.pl 
#			Initialize the modem from the current directory
#
#		> hsf_linstall.pl .
#			Initialize the modem from the current directory
#
#		> hsf_linstall.pl /modem
#			Initialize the modem from the directory /modem
#
#	Modify the country by country-name
#		> hsf_linstall.pl -c usa 
#			Set the country to USA
#
#		> hsf_linstall.pl -c japan
#			Set the country to JAPAN
#
#		> hsf_linstall.pl -c CANADA
#			Set the country to CANADA
#	
#	Modify the country by country-code
#		> hsf_linstall.pl -c 0xB5 
#			Set the country to USA
#		
#		> hsf_linstall.pl -c 0x0 
#			Set the country to JAPAN
#
#		> hsf_linstall.pl -c 0x00fe
#			Set the country to TAIWAN
#		
#	List the supported countries by the INF file
#		> hsf_linstall.pl -s
#
#	List the supported countries by the program and there codes
#		> hsf_linstall.pl -t
#	
#	Usage Information
#		> hsf_linstall.pl -u
#		> hsf_linstall.pl -usage
#	
#	Version information
#		> hsf_linstall.pl -v
#		> hsf_linstall.pl -version
#
#DIAGNOSTICS
#	WARNINGS
#		Specified current country code is not supported by program ('InfFilname')
#		Specified current country 'CurrentCountryName' is not supported by 'InfFilname'
#			The code which is configured in 'InfFilname' is not of one of the countries
#			that are supported by the program or by 'InfFilname'.
#			The most reasonable reason is that the INF file was configured by editing it
#			and not by using the -c option.
#			You should use the -s option to list the countries that the INF file support
#			and configure the INF file using the -c option.
#	
#	ERRORS
#		Can't open 'InstallDir' directory : 'error-message'	
#		Can't open PciFile : 'error-message'
#		Can't open   ... 'FileName' ...   : 'error-message'
#		Can't rename ... 'FileName' ...   : 'error-message'
#		Can't delete ... 'FileName' ...   : 'error-message'
#		Can't chdir to 'InstallDir' directory : 'error-message'
#		Can't chmod new created file :'error-message'
#			Make sure that you are running under root user
#			and that the file mode of 'PciFile' is correct
#			if yes, probably a system error has occurred.	
#
#		No SoftK56 modem PCI card found
#			No registration of Conexant PCI modem was found in the
#			PCI file (/proc/bus/pci/devices or /proc/pci).
#	
#		No matching INF file is found for SoftK56 PCI device
#			Matching INF file for the registered Conexant PCI modem	in the 
#			PciFile (/proc/bus/pci/devices or /proc/pci) was not found in the computer.
#
#		Application 'application-name' failed (returned ReturnCode),
#		Please view init_hsf.log 
#			Initialization failed. Viewing init_hsf.log might help in finding
#			the reason for Initialization failure.
#	
#EXIT CODES
#
#	0	No errors
#	
#	1	Usage or syntax error
#
#	2	Operational error or System error
#	
#REMARKS 
#	If no option were given the Initialization will take place, otherwise
#	the program will exit.
#	The configuration INF file and parameters to the program is NOT case sensitive.
#	
#AUTHOR
#	Conexant Systems Inc.
#
#COPYRIGHT
#	Copyright (C) 2000 Conexant Systems Inc.
#
#
__EOH__

use Getopt::Long;
#use strict "vars";
#use strict "refs";
#use strict "subs";

 
    # Initializations
$VERSION         = 'V3.0';
@PciFile         = ( '/proc/bus/pci/devices' , '/proc/pci' );
$InstallFilesExt = '\.inf';
$InitLogFile     = 'init_hsf.log';
 
%PciIds      = ();
$VendorID    = '';
$DeviceID    = '';
$InfFile     = '';
$debug       =  0;



    # Initialize Country Code Hash
    # Attention : Initialize the keys of the country name with UPPER CASE
%CountryCodes = ( 
    'USA'           => 0x00B5,
    'AUSTRIA'       => 0x000A,
    'BELGIUM'       => 0x000F,
    'CANADA'        => 0x0020,
    'CHINA'         => 0x0026,
    'CZECH'         => 0x002E,
    'DENMARK'       => 0x0031,
    'FINLAND'       => 0x003C,
    'FRANCE'        => 0x003D,
    'GERMANY'       => 0x0042,
    'GREECE'        => 0x0046,
    'HONG-KONG'     => 0x0050,
    'HUNGARY'       => 0x0051,
    'ICELAND'       => 0x0052,
    'INDIA'         => 0x0053,
    'INDONESIA'     => 0x0054,
    'IRELAND'       => 0x0057,
    'ISRAEL'        => 0x0058,
    'ITALY'         => 0x0059,
    'JAPAN'         => 0x0000,
    'KOREA'         => 0x0061,
    'LUXEMBOURG'    => 0x0069,
    'MALAYSIA'      => 0x006C,
    'MEXICO'        => 0x0073,
    'NETHERLAND'    => 0x007B,
    'NEW ZEALAND'   => 0x007E,
    'NORWAY'        => 0x0082,
    'PHILIPPINES'   => 0x0089,
    'PORTUGAL'      => 0x008B,
    'SINGAPORE'     => 0x009C,
    'SOUTH AFRICA'  => 0x009F,
    'SPAIN'         => 0x00A0,
    'SWEDEN'        => 0x00A5,
    'SWITZERLAND'   => 0x00A6,
    'TAIWAN'        => 0x00FE,
    'THAILAND'      => 0x00A9,
    'UK'            => 0x00B4,
    'AUSTRALIA'     => 0x0009,
);

    # Define hash for codes per country
%Code2Country;
while (my ($key, $value) = each %CountryCodes )
{
	$Code2Country{$value} = $key;
}




    # Canonize script name
($ScriptName = $0) =~ s,.*[\/\\],,;
# Do not put separate next two lines, or  Usage line will be wrong!!!
$InstallUsage = "Usage: $ScriptName [-c <country-name|country-code>] [-help] [-s]";
$InstallUsage .= "\n                       [-t] [-u|usage] [-v|version] [<release-dir>]\n";
# Do not put separate previous two lines, or Usage line will be wrong!!!


    # Get the flags of the command
$rc = GetOptions ('c=s' , 'help!', 's!', 't!', 'u|usage!', 'v|version!');
if (0 == $rc) 
{
	Error("$ScriptName: invalid option(s)\nTry `$ScriptName -help' for more information.");
}


    # Vesrion flag
if (defined $opt_help)
{
	PrintHelp();
	exit 0;
}



    # Vesrion flag
if (defined $opt_v)
{
	print "\n$ScriptName $VERSION\n";
	print "Copyright (C) 2000 Conexant Systems Inc.\n\n";
	exit 0;
}

    # Usage flag
if (defined $opt_u)
{
	print $InstallUsage;	
	exit 0;
}
    # Installation Supported countries flag
if (defined $opt_t)
{
	PrintInstallationCountries();
	exit 0;
}

    # Too many parameters
if (@ARGV > 1)

{
	Error("$ScriptName: too many parameters\nTry `$ScriptName -help' for more information.");
}

    # Install directory was given
if (@ARGV == 1) 
{
	$InstallDir = shift; 
}
    # Check install directory validity
$InstallDir = '.' unless defined $InstallDir;
Error("$ScriptName:  '$InstallDir': No such file or directory")
    unless -d $InstallDir;


    # Find out SoftK56 PCI modem
         # First try open /proc/bus/pci/devices
$index = 0;
if (open (PCI, "$PciFile[0]")) {
	while (<PCI>) {
		s/^\s*//o;
		# Example: 14f12016 (VendorId=14f1  DeviceId=2016)
		my ($device) = (split)[1];
		$device =~ /^([A-F0-9]{4})([A-F0-9]{4})$/io;
		$PciIds{uc "${1}_$index"} = uc $2;
		$index++;
    }
	close(PCI);
} else {
	     # In old kernel versions open the /proc/pci
	open (PCI, "$PciFile[1]") or
		Error("Can't open PciFile : $!");
	while (<PCI>) 
	{
        # Example: Vendor id=14f1. Device id=2015.
		next unless /^\s*Vendor\s*id\s*=\s*([a-f0-9]{1,4}).+Device\s*id\s*=\s*([a-f0-9]{1,4}).*$/io;
		$PciIds{uc "${1}_$index"} = uc $2;
		$index++;
	}
	close(PCI);
}

if ($debug) 
{
    for (keys %PciIds) 
    {
        printf "VendorID: %4s, DeviceID: %4s\n", $_, $PciIds{$_};
    }
}
Error("No SoftK56 modem PCI card found") unless %PciIds;

 
opendir(DIR, "$InstallDir") or 
	Error("Can't open '$InstallDir' directory : $!");
@InfFiles = grep(/$InstallFilesExt$/io, readdir(DIR));
close(DIR); 
print "Found INF files: @InfFiles\n" if $debug;
 
chdir $InstallDir or 
    Error("Can't chdir to '$InstallDir' directory : $!");


SEARCH_INF_FILES:
for my $file (@InfFiles) 
{
	open (INSTALLFILE, "$file") or
		Warning("Can't open '$file' file : $!\n"), next;

	while (<INSTALLFILE>) 
	{
            # Example; PCI\VEN_14F1&DEV_2016&SUBSYS_201614F1
		next unless /^.*PCI\\VEN_([a-f0-9]+)&DEV_([a-f0-9]+)(&SUBSYS_[a-f0-9]{8})?/io;
		$FileVendor = $1;
		$FileDeviceId = $2;
		for my $vendor (keys %PciIds)
		{
			$VendorKey = $vendor;
			$vendor =~ s/_.*//io;
            if ($vendor eq uc $FileVendor and $PciIds{$VendorKey} eq uc $FileDeviceId)
            {
				$InfFile  = $file;
				$VendorID = $vendor;
				$DeviceID = $PciIds{$VendorKey};
				print "Matching INF file is '$InfFile'\n" if $debug;
				close(INSTALLFILE);
				last SEARCH_INF_FILES;
			}
		}
	}	
	close(INSTALLFILE);
}
Error("No matching INF file is found for SoftK56 PCI device") unless $InfFile;




   # Print supported countries flag
if (defined $opt_s)
{
	PrintSupportedCountries();
	exit 0;
}

    # Change current-country flag
if (defined $opt_c)
{
	my $ctry = $opt_c;
	my $CurrCountry;

	   # Check for input (code or country)

	if ($ctry =~  m/^0x[a-f0-9]+$/io) # Code parameter
	{
		$ctry =~ s/^0X/0x/o; # Make 0x format instead of 0X
		my $octCtry = oct($ctry);

		   # Check code validity for the country-code
		if ( $octCtry  > oct('0xFFFF') ) 
		{ 
			Error ("Specified country code must be between 0x0 to 0xFFFF");
		}

		# Find the country by the code
		if ( exists $Code2Country{$octCtry} )
		{
			$CurrCountry = uc $Code2Country{$octCtry}
		}
		else
		{
			$ctry = sprintf "0x%04X" , $octCtry; 
			Error ("Code '$ctry' doesn't have supported country");
		}
	}
	else # Country name parameter
	{
		$CurrCountry = uc $opt_c;
	}


	Error("$ScriptName: Specified country '$CurrCountry' is not supported by installation")
		unless exists $CountryCodes{ $CurrCountry };

	    # The Inf File does not support the Current Country
    if ( 0 == SearchCurrentCountry($CurrCountry))
	{
		Error ("'$InfFile' does not support $CurrCountry country");
	}   
    # else ...
	ChangeCurrentCountry($CurrCountry);

	exit 0;
}
InstallModem($InfFile, $VendorID, $DeviceID);


###############################################################################
# SUB FUNCTIONS
###############################################################################

###############################################################################
    # Function prints all the supported countries in the installation.
    #    ( the countries that %CountryCodes ).

sub PrintInstallationCountries
{
    foreach $key (sort(keys %CountryCodes)) {
        printf "%-18s  0x%04X\n" , $key ,  $CountryCodes{$key};
    }
}

###############################################################################
    # Function prints all the countries that have support in the INF file,
    # In addition it markes the Current country.

sub PrintSupportedCountries
{
	my $line = '';
	my @country_arr;
	my $octCtry;
	my $currCountry;
	my $found = 0;
	
	open(ORGINF, $InfFile) or
		Error("Can't open '$InfFile' file : $!");


	    # Find the current country
 SEARCH_CURRENT_COUNTRY:
	while (! eof(ORGINF))
	{
		$line .= <ORGINF>;
		$line =~ s/\s*$//io;        # Remove trailing whitespaces
		if ($line =~ /\\$/io)       # Is this a continuation line?
		{
			chop $line;             # Remove \
			$line =~ s/\s*$//io;    # Remove possible whilespaces before \
			redo SEARCH_CURRENT_COUNTRY; # Continue...
		}
		else
		{
			$line =~ s/\s*//iog;        # Remove whitespaces
			    # Example : HKR, Country, Current, 1, 00,00
			if ($line =~  m/^\s*HKR\s*,\s*COUNTRY\s*,\s*CURRENT\s*,\s*[0-9a-f]{1,}\s*,\s*([0-9a-f]{2})\s*,\s*([0-9a-f]{2})\s*$/io)
			{          
				$line =~  s/^.*CURRENT\s*,\s*[0-9a-f]{1,}\s*,//io;
				$line =~  s/,//io;
			 	    # Example : 'B500'
				if ($line =~  m/^[0-9a-f]{4}$/io)
				{       # Create the hexa string
					$line = sprintf "0x%s%s" , substr($line,2,2) , substr($line,0,2);
					$octCtry = oct($line);
					if ( exists $Code2Country{$octCtry} )
					{
						$currCountry =  $Code2Country{$octCtry}
					}
				}
				last SEARCH_CURRENT_COUNTRY;
			}
			$line = '';
		}
	}
	
	    # Print the countries which are included in the INF file
	seek(ORGINF,0,0);
  SEARCH_COUNTRIES:
	while (! eof(ORGINF))
	{
		$line .= <ORGINF>;
		$line =~ s/\s*$//io;        # Remove trailing whitespaces
		if ($line =~ /\\$/io)       # Is this a continuation line?
		{
			chop $line;            # Remove \
			$line =~ s/\s*$//io;    # Remove possible whilespaces before \
			redo SEARCH_COUNTRIES; # Continue...
		}
		else
		{
			    # Example : HKR,Country\0000,NAME,,"USA"
			if ($line =~  m/^\s*HKR\s*,\s*COUNTRY\\([a-f0-9]{1,4}),\s*NAME\s*,/io)
			{          
				$line =~  s/.*NAME\s*,.*,\"//io;
				$line =~  s/\s*\"//io;
				$index = push @country_arr,uc $line;
			}
			$line = '';
		}
	}
	close(ORGINF);

	    # Print the sorted country array
	foreach my $ctry (sort(@country_arr)) 
	{
		{
			printf "%-15s" , $ctry;
			if ((defined $currCountry) && ( uc  $ctry eq uc $currCountry))
			{
				print "     (current country)";
				$found = 1;
			}
			print "\n";
		}
	}


	if (! defined $currCountry)
	{
		my $text = "Specified current country code is not supported by program ('$InfFile')";
		Warning( $text ) ;
	}
	elsif (0 == $found)
	{
		my $text = "Specified current country '$currCountry' is not supported by '$InfFile'";
		Warning( $text ) ;
	}

}

###############################################################################
    # Function modifies the current country-code in the Inf file to 
    #    the code of $CurrCountry

sub ChangeCurrentCountry
{
	my $CurrCountry = shift;
	my $line = '';
	my $newUpdateInf = $InfFile . '.temp_update';
	my $num = 0;


	open(ORGINF, $InfFile) or
		Error("Can't open '$InfFile' file : $!");
	if ( 0 == open(UPDATE_INF, ">$newUpdateInf"))
	{
		close (ORGINF);
		Error("Can't open temp file : $newUpdateInf $!"); 
	}
	    # System command failed
	$num = (chmod 0777 , $newUpdateInf);
	if ( 1 != $num )
	{
		close (ORGINF);
		close (UPDATE_INF);
		unlink ($newUpdateInf);
		Error ("Can't chmod new created file : $!");
	}


  COPY_INF_FILE:
	while (! eof(ORGINF))
	{
		$line .= <ORGINF>;

		    # Check if we got the Current Country line
		    # Example : HKR, Country, Current, 1, B5,00
		if ($line =~  m/^\s*HKR\s*,\s*COUNTRY,\s*CURRENT\s*,/io)
		{
			while ($line =~ /\\$/o)    # Is this a continuation line?
			{
				$line =~ s/\s*$//o;    # Remove trailing whitespaces
				chop $line;            # Remove \
				$line .= <ORGINF>;
			}
			$line =~ s/([0-9a-f]{2})\s*,\s*([0-9a-f]{2})\s*$//io;
			chop $line if $line =~ /\n$/io;      # Remove \n
			printf UPDATE_INF  "%s" ,$line;
			printf UPDATE_INF  "%02X"    , 0x00FF & $CountryCodes{ $CurrCountry };
			printf UPDATE_INF  ",%02X\n" , 0xFF00 & $CountryCodes{ $CurrCountry };
		}
		else
		{
			printf UPDATE_INF "%s" ,$line;
		}
		$line = '';
	}
	
        # Replace the Inf file to the Updated Inf file
	close ( ORGINF );
	close ( UPDATE_INF );
	if (0 == (rename $InfFile , "temp_file.tmp" )) # Rename failed
	{
		unlink $newUpdateInf;
		Error ("Can't rename $InfFile : $!");
	}
	if (0 == (rename $newUpdateInf , $InfFile))    # Rename failed
	{
		rename "temp_file.tmp" , $InfFile;
		unlink  $newUpdateInf;
		Error ("Can't rename '$newUpdateInf' (may be file was damaged): $!");
	}
	if (0 == (unlink "temp_file.tmp"))
	{
		Warning ("Can't delete temp file temp_file.tmp: $!\n");
	}
}


###############################################################################
    # Function Searches for the required Country in the Inf File.

sub SearchCurrentCountry
{
	my $CurrCountry = shift;
	my $line = '';



	open(ORGINF, $InfFile) or
		Error("Can't open '$InfFile' file : $!");

  SEARCH_COUNTRY:
	while (! eof(ORGINF))
	{
		$line .= <ORGINF>;
		$line =~ s/\s*$//o;        # Remove trailing whitespaces
		if ($line =~ /\\$/o)       # Is this a continuation line?
		{
			chop $line;            # Remove \
			$line =~ s/\s*$//o;    # Remove possible whilespaces before \
			redo SEARCH_COUNTRY;    # Continue...
		}
		else
		{
			# Example : HKR,Country\0000,NAME,,"USA"
			if (($line =~  m/^\s*HKR\s*,\s*COUNTRY\\([a-f0-9]{1,4}),\s*NAME\s*,/io)
				&&($line =~  m/,\s*\"$CurrCountry\"\s*/io))
			{                    
				# Country is supported in the Inf File
				return 1; 
			}
    		$line = '';
		}
	}
	close(ORGINF);
	return 0; # Country is not supported in the Inf File
}


###############################################################################
    # Function calls the ProcessInfFile that creates the binary Inf file.

sub InstallModem 
{
	my ($installFile, $vendor, $device) = @_;
	
	ProcessInfFile($installFile, $vendor, $device);
	
	return 1;
}

###############################################################################
    # Function creates the binary Inf file from the the Inf text file
    #    by the inf2bin application.

sub ProcessInfFile 
{
	my ($infFile, $vendor, $device) = @_;
	my $updatedInfFile = $infFile . '.tmp';
	my $line = '';
	my $infParseApp = 'inf2bin';

	
	open(ORGINF, $infFile) or
		Error("Can't open '$infFile' file : $!");
    open(NEWINF, ">$updatedInfFile") or
		Error("Can't open '$updatedInfFile' file: $!");
	print NEWINF "HKR,ID,VENDOR,,\"$vendor\"\n";
	print NEWINF "HKR,ID,DEVICE,,\"$device\"\n";

	
  READ_INF_FILE:
	while (! eof(ORGINF))
	{
		$line .= <ORGINF>;
		$line =~ s/;.*$//o;        # Remove possible comments
		$line =~ s/\s*$//o;        # Remove trailing whitespaces
        $line =~ s/^\s*//o;        # Remove initial whitespaces
        next if $line =~ /^\s*$/o; # Ignore empty lines
        $line =~ s/\s*,\s*/,/go;   # Remove all spaces around commas
		if ($line =~ /\\$/o)       # Is this a continuation line?
		{
			chop $line;            # Remove \
			$line =~ s/\s*$//o;    # Remove possible whilespaces before \
			redo READ_INF_FILE;    # Continue...
		}
		else
		{
            # Ignore non registry entries
            $line = '', next unless $line =~ /^HKR,/io;
			print NEWINF "$line\n";
			$line = '';
		}
	}
	close(ORGINF);
	close(NEWINF);
	

	my $command = "./$infParseApp $updatedInfFile /etc/SoftK56/bin_inf.dat >  $InitLogFile";
	print "$command\n" if $debug;
    $inf2bin_ret = system($command);
	if  ($inf2bin_ret)
	{
		unlink($updatedInfFile);
		Error("Application '$infParseApp' failed (returned $inf2bin_ret)\nPlease view $InitLogFile ");
	}
	
	unlink($updatedInfFile) or 
        Warning("Can't remove '$updatedInfFile' file: $!\n");
}

###############################################################################
    # Error sub routine including printing the Usage - logical problem

sub Usage {
    my $message = shift;
 
    print "$message" if $message;
    print $InstallUsage;	

    exit 1;
}

###############################################################################
    # Error sub routine

sub Error {
    my $message = shift;
 
    print "$ScriptName (error): $message\n";
 
    exit  2;
}
 

###############################################################################
    # Warning sub routine

sub Warning {
    my $message = shift;
 
    print "$ScriptName (warning): $message\n";
}

###############################################################################
    # Print the help for the help flag 

sub PrintHelp {
	$HelpText =~ s/#( )?//go;
	print $HelpText;
}
###############################################################################
