/********************************************************************************

File Name: HashTable.h

File Description:

********************************************************************************/

#ifndef __HASH_TABLE_H__
#define __HASH_TABLE_H__

#include "comtypes.h"
#include "linkedlist_ex.h"

#if ( USE_PRAGMA_PACK == 1 )
#pragma pack(push,1)   // Packing compatible with WinAC
#endif

#ifdef __cplusplus
extern "C"
{
#endif

typedef struct HASH_TABLE_TAG
{
	PLINKED_LIST_T	pHashLLTable;	/*	The hash table. Consists of handles  
										to linked lists */
	int				HashSize ;		/* Size of hash table */

	PALLOC_FUNC		pAllocFunc;
	PFREE_FUNC		pFreeFunc;
	PVOID			pRefData;		/* Reference Data pointer to be passed to pAllocFunc and pFreeFunc */
} HASH_TABLE_T, *PHASH_TABLE_T;

		
	/********************************************************************************

	Function Name: HTCreate
  
	Function Description: The function allocates a hash table instance, and constructs it

	Arguments:	TableSize - size of hash table
				pAllocFunc - a pointer to allocation function (usually, the memory manager Alloc function).
				pFreeWrap - a pointer to the free function.
				pRefData - a pointer to a reference data.

	Return Value: a pointer to the created hash table.

	********************************************************************************/
	PHASH_TABLE_T HT_Create(	IN	int				NumOfNodes, 
								IN	PALLOC_FUNC		pAllocWrap,
								IN	PFREE_FUNC		pFreeWrap,
								IN	PVOID			pRefData);

	/********************************************************************************

	Function Name: HT_Construct
  
	Function Description: The function initializes a hashtable instance fields without 
						allocations

	Arguments:	pHashTable
				TableSize - size of hash table
				pAllocFunc - a pointer to allocation function (usually, the memory manager Alloc function).
				pFreeWrap - a pointer to the free function.
				pRefData - a pointer to a reference data.

	Return Value: status indicating function success

	********************************************************************************/
	COM_STATUS HT_Construct(	IN	PHASH_TABLE_T	pHashTable,
								IN	int				NumOfNodes, 
								IN	PALLOC_FUNC		pAllocWrap,
								IN	PFREE_FUNC		pFreeWrap,
								IN	PVOID			pRefData);
	/********************************************************************************

	Function Name: HT_Init
  
	Function Description: The function initializes a hashtable instance with required 
							allocations and member's create which will be undone by 
							the close function.

	Arguments: pHashTable - HashTable instance .

	Return Value: None.

	********************************************************************************/
	COM_STATUS HT_Init(			IN	PHASH_TABLE_T	pHashTable);
	/********************************************************************************

	Function Name: HT_Close
  
	Function Description: The function undoes all actions done by init.

	Arguments: pHashTable - HashTable instance .

	Return Value: None.

	********************************************************************************/
	COM_STATUS HT_Close(			IN	PHASH_TABLE_T	pHashTable);
	/********************************************************************************

	Function Name: HTDestroy
  
	Function Description: The function destroys the HashTable parameters.

	Arguments: pHashTable - a handle to the HashTable instance (this).
   			   pFreeWrap - a pointer to the free function.

	Return Value: None.

	********************************************************************************/
	void HT_Destroy(			IN PHASH_TABLE_T pHashTable);

	/********************************************************************************

	Function Name: HT_Insert
  
	Function Description: The function inserts a new node to the HashTable according 
						  to its Id.

	Arguments:	pHashTable - HashTable instance (this).
				Id - Id for the new node
				pNode  - a pointer to the new node we want to insert.

	Return Value: None.

	********************************************************************************/
	BOOL HT_Insert(			IN	PHASH_TABLE_T	pHashTable, 
							IN	UINT32			Id,
							IN	PVOID			pData);

	/********************************************************************************


	Function Name: HT_Remove.
  
	Function Description: The function deletes the node with the identifier Id 
						  from the HashTable.

	Arguments: pHashTable - HashTable instance (this).
			   Id - the identifier of the node we want to delete.

	Return Value: None.

	********************************************************************************/
	void HT_Remove(		IN	PHASH_TABLE_T	pHashTable, 
						IN	UINT32			Id,
						OUT	PVOID*			ppData	);


	/********************************************************************************

	Function Name: HT_SearchNode
  
	Function Description: The function search a node in the HashTable with the identifier
						  Id.

	Arguments: pHashTable - HashTable instance (this).
			   Id - the identifier of the node we want to search.
			   pData - a pointer to a VOID (to get the pData value of the node).

	Return Value: returns TRUE when node with Id was found, 
					FALSE otherwise.

	********************************************************************************/
	BOOL HT_SearchNode(		IN	PHASH_TABLE_T	pHashTable, 
							IN	UINT32			Id, 
							OUT	PVOID			*pData);

	/********************************************************************************

	Function Name: HT_GetFirstNode
  
	Function Description: This function returnes the first node in the 
			hash table data base.

	Arguments: pHashTable - HashTable instance (this).
			   pId - Pointer to Id to hold the first node ID.
			   ppData - a pointer to a VOID (to 
			   get the pData value of the node).

	Return Value: Returns TRUE when there is a node in the hash, false when
					the hash is empty.

	********************************************************************************/
	BOOL HT_GetFirstNode(	IN	PHASH_TABLE_T	pHashTable, 
							OUT	UINT32			*pId,	
							OUT	PVOID			*ppData);

#ifdef __cplusplus
}
#endif

#if ( USE_PRAGMA_PACK == 1 )
#pragma pack(pop)   // Packing compatible with our VXDs
#endif

#endif /*__HASH_TABLE_H__*/
