/********************************************************************************

File Name: LinkedList.h

File Description:

********************************************************************************/

#ifndef __LINKED_LIST_H__
#define __LINKED_LIST_H__

#include "comtypes.h"

#if ( USE_PRAGMA_PACK == 1 )
#pragma pack(push,1)   // Packing compatible with WinAC
#endif

#ifdef __cplusplus
extern "C"
{
#endif

typedef enum { ORDER_TYPE, NON_ORDER_TYPE, } LINKED_LIST_TYPE;
typedef enum { AFTER, BEFORE, HEAD, TAIL, } INSERT_TYPE;

typedef struct LLNODE_TAG
{
	UINT32				Id;
	PVOID				pData;
	struct LLNODE_TAG *	pNext;
	struct LLNODE_TAG *	pPrev;
} LLNODE_T, *PLLNODE_T;

typedef struct LINKED_LIST_TAG
{
	PLLNODE_T			pHead;		 	/* a pointer to the first node of the linked list. */
	PLLNODE_T			pTail;		 	/* a pointer to the last node of the linked list. */
	LINKED_LIST_TYPE	ListType;	 	/* the linked list type. default: order according to the index. */
	PALLOC_FUNC			pAllocFunc;
	PFREE_FUNC			pFreeFunc;
	HANDLE				hCriticalSection;
	PVOID				pRefData;		/* Reference Data pointer to be passed to pAllocFunc and pFreeFunc */
} LINKED_LIST_T, *PLINKED_LIST_T;

		
	/********************************************************************************

	Function Name: LL_CreateLinkedList
  
	Function Description: The function inits the LinkedList parameters.

	Arguments: Type - the list type (ordered or not).
			   pAllocFunc - a pointer to allocation function (usually, the memory manager Alloc function).
			   pFreeWrap - a pointer to the free function.
			   pRefData - a pointer to a reference data.

	Return Value: a handle to the created linked list.

	********************************************************************************/
	PLINKED_LIST_T LL_CreateLinkedList	(	IN	LINKED_LIST_TYPE	Type,
											IN	PALLOC_FUNC			pAllocWrap,
											IN	PFREE_FUNC			pFreeWrap, 
											IN	PVOID				pRefData);

	/********************************************************************************

	Function Name: LL_ConstructLinkedList
  
	Function Description: The function inits the LinkedList parameters.

	Arguments: pLinkedList - a pointer to the linkedlist instance.
			   Type - the list type (ordered or not).
			   pAllocFunc - a pointer to allocation function (usually, the memory manager Alloc function).
			   pFreeWrap - a pointer to the free function.
			   pRefData - a pointer to a reference data.

	Return Value: a handle to the created linked list.

	********************************************************************************/
	COM_STATUS LL_ConstructLinkedList(	IN	PLINKED_LIST_T		pLinkedList,
										IN	LINKED_LIST_TYPE	Type,
										IN	PALLOC_FUNC			pAllocWrap,
										IN	PFREE_FUNC			pFreeWrap, 
										IN	PVOID				pRefData);

	/********************************************************************************

	Function Name: LL_ConstructLinkedList
  
	Function Description: The function inits the LinkedList parameters.

	Arguments: pLinkedList - a pointer to the linkedlist instance.
			   Type - the list type (ordered or not).
			   pAllocFunc - a pointer to allocation function (usually, the memory manager Alloc function).
			   pFreeWrap - a pointer to the free function.
			   pRefData - a pointer to a reference data.

	Return Value: a handle to the created linked list.

	********************************************************************************/
	COM_STATUS LL_ConstructLinkedList(	IN	PLINKED_LIST_T		pLinkedList,
										IN	LINKED_LIST_TYPE	Type,
										IN	PALLOC_FUNC			pAllocWrap,
										IN	PFREE_FUNC			pFreeWrap, 
										IN	PVOID				pRefData);
	/********************************************************************************

	Function Name: LL_DestroyLinkedList
  
	Function Description: The function destroys the LinkedList parameters.

	Arguments: hLinkedList - a handle to the LinkedList instance (this).
   			   pFreeWrap - a pointer to the free function.

	Return Value: None.

	********************************************************************************/
	void LL_DestroyLinkedList(IN PLINKED_LIST_T pLinkedList);

	void LL_DestructLinkedList(IN PLINKED_LIST_T pLinkedList);

	/********************************************************************************

	Function Name: LL_CreateNode
  
	Function Description:	The function creates a node handle that can be 
							used by a linked list.

	Arguments:	pFuncAlloc	-	function for needed allocations
				pRefData	-	Reference Data for pFuncAlloc.


	Return Value: handle to the node created.

	********************************************************************************/
	HANDLE LL_CreateNode(	IN	PALLOC_FUNC		pFuncAlloc,
							IN	PVOID			pRefData);
	/********************************************************************************

	Function Name: LL_DestroyNode
  
	Function Description:	The function destroys a node. The node should 
							be removed from linked list before this function
							is called.

	Arguments:	hNode		-	Handle to the node to be destroyed.
				pFuncFree	-	function for needed deallocations
				pRefData	-	Reference Data for pFuncFree.


	Return Value: None.

	********************************************************************************/
	void LL_DestroyNode(	IN	HANDLE			hNode,
							IN	PFREE_FUNC		pFuncFree,
							IN	PVOID			pRefData);
	/********************************************************************************

	Function Name: LL_InsertOrder
  
	Function Description: The function inserts a new node to the LinkedList according 
						  to its Id.

	Arguments:	hLinkedList	-	a handle to the LinkedList instance (this).
				hNode		-	Handle to node to store the data. If this 
								parameter is NULL, the function creates a 
								node.


	Return Value: None.

	********************************************************************************/
	BOOL LL_InsertOrder(	IN	PLINKED_LIST_T	pLinkedList,
							IN	UINT32			Id, 
							IN	PVOID			pData,
							IN	HANDLE			hNode );

	/********************************************************************************

	Function Name: LL_InsertNonOrder
  
	Function Description: The function inserts a new node to the LinkedList according 
						  to its Id.

	Arguments:	hLinkedList	-	a handle to the LinkedList instance (this).
				Id			-	the id of the new node.
				pData		-	the data of the new node.
				hInputNode	-	a pointer to the node that should relative to the new 
								node (as chosen in the InsertType parameter).
				InsertType	-	The type of the insert (AFTER, BEFORE etc).
				hNode		-	Handle to node to store the data. If this 
								parameter is NULL, the function creates a 
								node.

	Return Value: None.

	********************************************************************************/
	BOOL LL_InsertNonOrder(	IN	PLINKED_LIST_T	pLinkedList, 
							IN	UINT32			Id, 
							IN	PVOID			pData, 
							IN	HANDLE			hInputNode, 
							IN	INSERT_TYPE		InsertType,
							IN	HANDLE			hNode);

	/********************************************************************************

	Function Name: LL_RemoveById.
  
	Function Description: The function removes the node with the identifier Id 
						  from the LinkedList.

	Arguments:	hLinkedList	-	a handle to the LinkedList instance (this).
				Id			-	the identifier of the node we want to delete.
				ppData		-	the pData pointer will retuen in ppData parameter.
								(the ppData parameter can be NULL, if pData isn't necessary).
				bDestroyNode-	indicates if node should be destroyed by this function

	Return Value: handle of node removed. if bDestroyNode parameter is TRUE, returns NULL

	********************************************************************************/
	HANDLE LL_RemoveById(	IN	PLINKED_LIST_T	pLinkedList, 
							IN	UINT32			Id,
							OUT	PVOID			* ppData,
							IN	BOOL			bDestroyNode );

	/********************************************************************************

	Function Name: LL_RemoveNode.
  
	Function Description: The function removes the node with the identifier Id 
						  from the LinkedList.

	Arguments: hLinkedList - a handle to the LinkedList instance (this).
			   pNode - the node to delete.
			   ppData - the pData pointer will retuen in ppData parameter.
						(the ppData parameter can be NULL, if pData isn't necessary).
				bDestroyNode-	indicates if node should be destroyed by this function
	Return Value: None

	********************************************************************************/
	void LL_RemoveNode(	IN	PLINKED_LIST_T	pLinkedList,
						IN	HANDLE			hNode,
						OUT	PVOID			* ppData,
						IN	BOOL			bDestroyNode);

	/********************************************************************************

	Function Name: LL_SearchNode
  
	Function Description: The function search a node in the LinkedList with the identifier
						  Id.

	Arguments: hLinkedList - a handle to the LinkedList instance (this).
			   Id - the identifier of th enode we want to search.
			   pData - a pointer to a VOID (to get the pData value of the node).

	Return Value: A handle to the node with Id as its identifier.

	********************************************************************************/
	HANDLE LL_SearchNode(IN PLINKED_LIST_T pLinkedList, IN UINT32 Id, OUT PVOID* pData);


	/********************************************************************************

	Function Name: LL_GetFirstNode
  
	Function Description: The function  returns the first node of the specific Linked-List
						  
	Arguments: hLinkedList - a handle to the LinkedList instance (this).
			   pId - a pointer to a UINT32 (to get the Id value of the first node).
			   pData - a pointer to a VOID (to get the pData value of the first node).

	Return Value: A handle of the first node of the specific Linked-List (which pHead is
	pointing to).

	********************************************************************************/
	HANDLE LL_GetFirstNode(IN PLINKED_LIST_T pLinkedList,OUT UINT32* pId, OUT PVOID* pData);

	/********************************************************************************

	Function Name: LL_GetNextNode
  
	Function Description: The function  returns the next node of the specific Linked-List.
						  There is a pointer to the current node in the LinkedList (that is 
						  updated only by using the functions: GetFirstNode and GetNextNode) and the 
						  function return the node that is after the current node in the 
						  specific linked-list.

	Arguments: hLinkedList - a handle to the LinkedList instance (this).
			   hNode - handle of the last node (returned by GetFirstNode or GetNextNode).
   			   pId - a pointer to a UINT32 (to get the Id value of the first node).
			   pData - a pointer to a VOID (to get the pData value of the first node).

	Return Value: A handle of the next node of the specific Linked-List

	********************************************************************************/
	HANDLE LL_GetNextNode(IN PLINKED_LIST_T pLinkedList, IN HANDLE hNode, OUT UINT32* pId, OUT PVOID* pData);

#ifdef __cplusplus
}
#endif

#if ( USE_PRAGMA_PACK == 1 )
#pragma pack(pop)   // Packing compatible with our VXDs
#endif

#endif /*__LINKED_LIST_H__*/
