/*
 *  linux/drivers/char/serial.c
 *
 *  Copyright (C) 1991, 1992  Linus Torvalds
 *
 *  Extensively rewritten by Theodore Ts'o, 8/16/92 -- 9/14/92.  Now
 *  much more extensible to support other serial cards based on the
 *  16450/16550A UART's.  Added support for the AST FourPort and the
 *  Accent Async board.
 *
 *  set_serial_info fixed to set the flags, custom divisor, and uart
 * 	type fields.  Fix suggested by Michael K. Johnson 12/12/92.
 *
 *  11/95: TIOCMIWAIT, TIOCGICOUNT by Angelo Haritsis <ah@doc.ic.ac.uk>
 *
 *  03/96: Modularised by Angelo Haritsis <ah@doc.ic.ac.uk>
 *
 *  rs_set_termios fixed to look also for changes of the input
 *      flags INPCK, BRKINT, PARMRK, IGNPAR and IGNBRK.
 *                                            Bernd Anhupl 05/17/96.
 *
 *  1/97:  Extended dumb serial ports are a config option now.
 *         Saves 4k.   Michael A. Griffith <grif@acm.org>
 *
 *  8/97: Fix bug in rs_set_termios with RTS
 *        Stanislav V. Voronyi <stas@uanet.kharkov.ua>
 *
 *  3/98: Change the IRQ detection, use of probe_irq_o*(),
 *	  supress TIOCSERGWILD and TIOCSERSWILD
 *	  Etienne Lorrain <etienne.lorrain@ibm.net>
 *
 *  4/98: Added changes to support the ARM architecture proposed by
 * 	  Russell King
 *
 * This module exports the following rs232 io functions:
 *
 *	int rs_init(void);
 */

/*
 * Serial driver configuration section.  Here are the various options:
 *
 * CONFIG_HUB6
 *		Enables support for the venerable Bell Technologies
 *		HUB6 card.
 *
 * CONFIG_SERIAL_MANY_PORTS
 * 		Enables support for ports beyond the standard, stupid
 * 		COM 1/2/3/4.
 *
 * CONFIG_SERIAL_MULTIPORT
 * 		Enables support for special multiport board support.
 *
 * CONFIG_SERIAL_SHARE_IRQ
 * 		Enables support for multiple serial ports on one IRQ
 *
 * CONFIG_SERIAL_DETECT_IRQ
 *		Enable the autodetection of IRQ on standart ports
 *
 * SERIAL_PARANOIA_CHECK
 * 		Check the magic number for the async_structure where
 * 		ever possible.
 */

#define NOT_SKELETON_SERIAL

#undef SERIAL_PARANOIA_CHECK
#define SERIAL_DO_RESTART

#include "framewrk.h"

#if 0
/* Normally these defines are controlled by the autoconf.h */

#define CONFIG_SERIAL_MANY_PORTS
#define CONFIG_SERIAL_SHARE_IRQ
#define CONFIG_SERIAL_DETECT_IRQ
#define CONFIG_SERIAL_MULTIPORT
#define CONFIG_HUB6
#endif

/* Sanity checks */

#ifdef CONFIG_SERIAL_MULTIPORT
#ifndef CONFIG_SERIAL_SHARE_IRQ
#define CONFIG_SERIAL_SHARE_IRQ
#endif
#endif

/* Set of debugging defines */

#define SERIAL_DEBUG_INTR
#undef SERIAL_DEBUG_OPEN
#define SERIAL_DEBUG_FLOW
#define SERIAL_DEBUG_RS_WAIT_UNTIL_SENT

#define RS_STROBE_TIME (10*HZ)
#define RS_ISR_PASS_LIMIT 256

#define IRQ_T(info) ((info->flags & ASYNC_SHARE_IRQ) ? SA_SHIRQ : SA_INTERRUPT)

#define SERIAL_INLINE

#if defined(MODULE) && defined(SERIAL_DEBUG_MCOUNT)
#define DBG_CNT(s) printk("(%s): [%x] refc=%d, serc=%d, ttyc=%d -> %s\n", \
 kdevname(tty->device), (info->flags), serial_refcount,info->count,tty->count,s)
#else
#define DBG_CNT(s)
#endif

/*
 * End of serial driver configuration section.
 */

#include <linux/config.h>
#include <linux/module.h>
#include <linux/errno.h>
#include <linux/signal.h>
#include <linux/sched.h>
#include <linux/timer.h>
#include <linux/interrupt.h>
#include <linux/tty.h>
#include <linux/tty_flip.h>
#include <linux/serial.h>
#include <linux/serialP.h>
#include <linux/serial_reg.h>
#include <linux/major.h>
#include <linux/string.h>
#include <linux/fcntl.h>
#include <linux/ptrace.h>
#include <linux/ioport.h>
#include <linux/mm.h>
#include <linux/malloc.h>
#include <linux/init.h>
#include <linux/delay.h>

#include <asm/system.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <asm/uaccess.h>
#include <asm/bitops.h>
#include <asm/serial.h>
#include <linux/tqueue.h>

/* ignore __stdcall */
#define __stdcall

#include "ossysenv.h"
#include "ostime_ex.h"
#ifndef SKELETON_SERIAL
#include "intfctrl_ex.h"
#include "comctrl_ex.h"
#include "port.h"
#endif

#ifdef SERIAL_INLINE
#define _INLINE_ inline
#endif

static char *serial_name = "Conexant Serial Driver";
static char *serial_version = "1.00";

static DECLARE_TASK_QUEUE(tq_serial);
//static DECLARE_TASK_QUEUE(tq_cnxt);

static struct tq_struct cnxt_task_struct;
void cnxt_task(void *data);

static struct tty_driver serial_driver, callout_driver;
static int serial_refcount;

static struct timer_list serial_timer;

/* number of characters left in xmit buffer before we ask for more */
#define WAKEUP_CHARS 256

/*
 * IRQ_timeout		- How long the timeout should be for each IRQ
 * 				should be after the IRQ has been active.
 */

static struct async_struct *IRQ_ports[NR_IRQS];
static int IRQ_timeout[NR_IRQS];

static void autoconfig(struct serial_state * info);
static void change_speed(struct async_struct *info, struct termios *old);
static void rs_wait_until_sent(struct tty_struct *tty, int timeout);


/*
 * Here we define the default xmit fifo size used for each type of
 * UART
 */
static struct serial_uart_config uart_config[] = {
	{ "unknown", 1, 0 },
	{ "8250", 1, 0 },
	{ "16450", 1, 0 },
	{ "16550", 1, 0 },
	{ "16550A", 16, UART_CLEAR_FIFO | UART_USE_FIFO },
	{ "cirrus", 1, 0 },
	{ "ST16650", 1, UART_CLEAR_FIFO | UART_STARTECH },
	{ "ST16650V2", 32, UART_CLEAR_FIFO | UART_USE_FIFO |
		  UART_STARTECH },
	{ "TI16750", 64, UART_CLEAR_FIFO | UART_USE_FIFO},
	{ 0, 0}
};

static struct serial_state rs_table[] = {
	SERIAL_PORT_DFNS	/* Defined in serial.h */
};

static struct {
	UINT32	ier;
	UINT32	lcr;
	UINT32	mcr;
	UINT32	lsr;
	UINT32	msr;
} uart_shadow = { 0, 0, 0, 0, 0 };

#define NR_PORTS	(sizeof(rs_table)/sizeof(struct serial_state))

static struct tty_struct *serial_table[NR_PORTS];
static struct termios *serial_termios[NR_PORTS];
static struct termios *serial_termios_locked[NR_PORTS];

#ifndef MIN
#define MIN(a,b)	((a) < (b) ? (a) : (b))
#endif

/*
 * tmp_buf is used as a temporary buffer by serial_write.  We need to
 * lock it in case the copy_from_user blocks while swapping in a page,
 * and some other program tries to do a serial write at the same time.
 * Since the lock will only come under contention when the system is
 * swapping and available memory is low, it makes sense to share one
 * buffer across all the serial ports, since it significantly saves
 * memory if large numbers of serial ports are open.
 */
static unsigned char *tmp_buf = NULL;

#if ( LINUX_VERSION_CODE < KERNEL_VERSION( 2, 4, 0 ) )
static struct semaphore tmp_buf_sem = MUTEX;

#define XMIT_HEAD xmit_head
#define XMIT_TAIL xmit_tail
#define XMIT_BUF  xmit_buf

/* Return space available up to the end of the buffer.  */
#define CIRC_SPACE_TO_END(head,tail,size) \
        ({int end = (size) - 1 - (head); \
          int n = (end + (tail)) & ((size)-1); \
          n <= end ? n : end+1;})

#define CIRC_CNT(head,tail,size) (((head) - (tail)) & ((size)-1))
#define CIRC_SPACE(head,tail,size) CIRC_CNT((tail),((head)+1),(size))

#else // kernel version >= 2.4.0
static DECLARE_MUTEX(tmp_buf_sem);

#define XMIT_HEAD xmit.head
#define XMIT_TAIL xmit.tail
#define XMIT_BUF  xmit.buf
#endif

/* Shay: below is stuff that was ripped from octopus' PortDrv */
//void rs_eventNotify(HANDLE hport, UINT32 dwEvtMask);
int init_com_if( void );

#ifndef SKELETON_SERIAL
PORTINFORMATION ExtendedInfo;
#endif

static inline int serial_paranoia_check(struct async_struct *info,
					kdev_t device, const char *routine)
{
#ifdef SERIAL_PARANOIA_CHECK
	static const char *badmagic =
		"Warning: bad magic number for serial struct (%s) in %s\n";
	static const char *badinfo =
		"Warning: null async_struct for (%s) in %s\n";

	if (!info) {
		printk(badinfo, kdevname(device), routine);
		return 1;
	}
	if (info->magic != SERIAL_MAGIC) {
		printk(badmagic, kdevname(device), routine);
		return 1;
	}
#endif
	return 0;
}

static inline unsigned int serial_in(struct async_struct *info, int offset)
{
	unsigned int ret;
	unsigned long int flags;

	switch ( offset )
	{

	case UART_IER:
		return uart_shadow.ier;

	case UART_LCR:
		return uart_shadow.lcr;

	case UART_MCR:
		return uart_shadow.mcr;

	case UART_LSR:
		return uart_shadow.lsr;

	case UART_MSR:
		save_flags(flags); cli();
		ret = uart_shadow.msr;
		uart_shadow.msr &= ~(UART_MSR_DCTS | UART_MSR_DDSR |
							 UART_MSR_TERI | UART_MSR_DDCD);
		restore_flags(flags);
		return ret;

	}

	printk( "serial_in: unknown / unsupported register #%d\n", offset );
	return 0;

}

static inline int serial_out(struct async_struct *info, int offset, int value)
{

	switch ( offset )
	{

	case UART_IER:
		uart_shadow.ier = value;
		break;

	case UART_LCR:
		uart_shadow.lcr = value;
		break;

	case UART_MCR:
		uart_shadow.mcr = value;
		break;

	case UART_LSR:
		uart_shadow.lsr = value;
		break;

	case UART_MSR:
		uart_shadow.msr = value;
		break;

	case UART_TX:
#ifndef SKELETON_SERIAL
	{
		char cbyte = (char)value;

		if ( NULL != ExtendedInfo.pComCtrl )
		{
/*  		printk("info->port %x, %d\n",info->port,offset); */
			return ExtendedInfo.pComCtrl->Write(ExtendedInfo.hComCtrl,
												&cbyte, 1);
		}
		break;
	}
#endif

	default:
		return 0;
	}

	return 1;

}  /* serial_out */


/*
 * ------------------------------------------------------------
 * rs_stop() and rs_start()
 *
 * This routines are called before setting or resetting tty->stopped.
 * They enable or disable transmitter interrupts, as necessary.
 * ------------------------------------------------------------
 */
static void rs_stop(struct tty_struct *tty)
{
	struct async_struct *info = (struct async_struct *)tty->driver_data;
	/*unsigned long flags; */

	printk( "cnxt serial: rs_stop\n" );

	if (serial_paranoia_check(info, tty->device, "rs_stop"))
		return;
	/*
	save_flags(flags); cli();
	*/
	if (info->IER & UART_IER_THRI) {
		info->IER &= ~UART_IER_THRI;
	/*
		serial_out(info, UART_IER, info->IER);
	*/
	}
	/*
	restore_flags(flags);
	*/
}

static void rs_start(struct tty_struct *tty)
{
	struct async_struct *info = (struct async_struct *)tty->driver_data;
	/*unsigned long flags;*/

	printk( "cnxt serial: rs_start\n" );

	if (serial_paranoia_check(info, tty->device, "rs_start"))
		return;
	/*
	save_flags(flags); cli();
	*/

	if ( info->XMIT_HEAD != info->XMIT_TAIL &&
		 info->XMIT_BUF &&
		 !(info->IER & UART_IER_THRI) )
	{
		info->IER |= UART_IER_THRI;
		/* serial_out(info, UART_IER, info->IER); */
	}

	/*
	restore_flags(flags);
	*/
}

/*
 * ----------------------------------------------------------------------
 *
 * Here starts the interrupt handling routines.  All of the following
 * subroutines are declared as inline and are folded into
 * rs_interrupt().  They were separated out for readability's sake.
 *
 * Note: rs_interrupt() is a "fast" interrupt, which means that it
 * runs with interrupts turned off.  People who may want to modify
 * rs_interrupt() should try to keep the interrupt handler as fast as
 * possible.  After you are done making modifications, it is not a bad
 * idea to do:
 *
 * gcc -S -DKERNEL -Wall -Wstrict-prototypes -O6 -fomit-frame-pointer serial.c
 *
 * and look at the resulting assemble code in serial.s.
 *
 * 				- Ted Ts'o (tytso@mit.edu), 7-Mar-93
 * -----------------------------------------------------------------------
 */

/*
 * This routine is used by the interrupt handler to schedule
 * processing in the software interrupt portion of the driver.
 */
static _INLINE_ void rs_sched_event(struct async_struct *info,
				  int event)
{
	printk( "cnxt serial: rs_sched_event tqueue->routine = 0x%x\n",info->tqueue.routine );

	info->event |= 1 << event;
	queue_task(&info->tqueue, &tq_serial);
	mark_bh(SERIAL_BH);
}

/* received status is not valid until flags system exists fixed */
static _INLINE_ void receive_chars(struct async_struct *info, int *status)
{
	struct tty_struct *tty = info->tty;
	unsigned char ch;
//	int ignored = 0;
	struct	async_icount *icount;
	unsigned long	bytes_read;
	char            read_buf[TTY_FLIPBUF_SIZE];
	int i;

#ifndef SKELETON_SERIAL

	if ( NULL != ExtendedInfo.pComCtrl )
	{
		bytes_read = ExtendedInfo.pComCtrl->Read(ExtendedInfo.hComCtrl, (void*)read_buf, TTY_FLIPBUF_SIZE);
	}
	else
		return;
#else

	bytes_read = 0;

#endif

	icount = &info->state->icount;
	for ( i = 0; i < bytes_read; i++ ) {
		ch = read_buf[i];
		if (tty->flip.count >= TTY_FLIPBUF_SIZE)
			break;
//printk( "receive_chars: 1 byte\n" );
		*tty->flip.char_buf_ptr = ch;
		icount->rx++;

		*tty->flip.flag_buf_ptr = 0;
//		if (*status & (UART_LSR_BI | UART_LSR_PE | /* Shay: break, parity error, frame error, overrun error */
//			       UART_LSR_FE | UART_LSR_OE)) {
			/*
			 * For statistics only
			 */
//			if (*status & UART_LSR_BI) {
//				*status &= ~(UART_LSR_FE | UART_LSR_PE);
//				icount->brk++;
//			} else if (*status & UART_LSR_PE)
//				icount->parity++;
//			else if (*status & UART_LSR_FE)
//				icount->frame++;
//			if (*status & UART_LSR_OE)
//				icount->overrun++;

			/*
			 * Now check to see if character should be
			 * ignored, and mask off conditions which
			 * should be ignored.
			 */
//			if (*status & info->ignore_status_mask) {
//				if (++ignored > 100)
//					break;
//				goto ignore_char;
//			}
//			*status &= info->read_status_mask;

//			if (*status & (UART_LSR_BI)) {

//#ifdef SERIAL_DEBUG_INTR
//				//printk("handling break....");
//#endif
//				*tty->flip.flag_buf_ptr = TTY_BREAK;
//				if (info->flags & ASYNC_SAK)
//					do_SAK(tty);
//			} else if (*status & UART_LSR_PE) /* Shay: parity error */
//				*tty->flip.flag_buf_ptr = TTY_PARITY;
//			else if (*status & UART_LSR_FE)
//				*tty->flip.flag_buf_ptr = TTY_FRAME;
//			if (*status & UART_LSR_OE) {

				/*
				 * Overrun is special, since it's
				 * reported immediately, and doesn't
				 * affect the current character
				 */
//				if (tty->flip.count < TTY_FLIPBUF_SIZE) {
//					tty->flip.count++;
//					tty->flip.flag_buf_ptr++;
//					tty->flip.char_buf_ptr++;
//					*tty->flip.flag_buf_ptr = TTY_OVERRUN;
//				}
//			}
//		}
		tty->flip.flag_buf_ptr++;
		tty->flip.char_buf_ptr++;
		tty->flip.count++;
//	ignore_char:
//		*status = serial_in(info, UART_LSR);
	}

	tty_flip_buffer_push(tty);
}

/* move the bytes downstairs (atparser?) */
static _INLINE_ void transmit_chars(struct async_struct *info)
{
	unsigned long flags;

	while ( (info->XMIT_HEAD != info->XMIT_TAIL) &&
			( !C_CRTSCTS(info->tty) ||
			  ((serial_in(info, UART_MSR) & UART_MSR_CTS) != 0) ) )
	{
		if ( serial_out(info, UART_TX, info->XMIT_BUF[info->XMIT_TAIL]) != 1 )
			break;

		/* Actually, we should add the serial_out to the critical section,
		 * since it sends the char buf[info->xmit_tail], and xmit_tail might be
		 * changed. The problem is that this will run the atparser as critical,
		 * which does not seem right. I think it's not too bad, since:
		 * A: There was no critical section - I added it :-)
		 * B: The worst thing is that the wrong char will be sent. */
		save_flags(flags); cli();
		info->state->icount.tx++;
		if ( (++info->XMIT_TAIL) >= SERIAL_XMIT_SIZE )
			info->XMIT_TAIL = 0;
		restore_flags(flags);
	}

	// remove the UART_IER_THRI bit from the IER (Interrupt Enable Register) to keep
	// the info structure updated
	if (info->XMIT_HEAD == info->XMIT_TAIL) {
		info->IER &= ~UART_IER_THRI;
	}


//    if( info->xmit_cnt <= 0 )
//	   *pport->pEvtDword |= EV_TXEMPTY;
//    if( dwSize < dwInitialSize )
//	   *pport->pEvtDword |= EV_TXCHAR;

//    if( dwInitialSize >= pport->TxTrigger  &&
//        dwSize < pport->TxTrigger          &&
//        pport->pfnWriteNotifyHandle            )
//    {
//		pport->pfnWriteNotifyHandle( pport, pport->WriteNotifyRefData, CN_TRANSMIT, *pport->pEvtDword );
//    }

}

static _INLINE_ void check_modem_status(struct async_struct *info)
{
	static int counter = 0, wake_counter = -1;
	int status = serial_in(info, UART_MSR);

	counter++;
	/* Make sure we don't overlap counter: */
	if ( (counter > 1000000) && (wake_counter == -1) )
		counter = 0;

#if 0
	printk( "cms: status = 0x%x (DDCD=%d,DCD=%d,CHECK_CD=%d)\n",
			status,
			(status & UART_MSR_DDCD) ? 1 : 0,
			(status & UART_MSR_DCD) ? 1 : 0,
			(info->flags & ASYNC_CHECK_CD) ? 1 : 0 );
#endif

    if ( (info->flags & ASYNC_CHECK_CD) && (status & UART_MSR_DDCD)) {
		printk("ttys%d CD now %s...", info->line,
		       (status & UART_MSR_DCD) ? "on" : "off");

		if (status & UART_MSR_DCD) {
			printk( "blocking a bit more...\n" );
			wake_counter = counter;
		}
		else if (!((info->flags & ASYNC_CALLOUT_ACTIVE) &&
				   (info->flags & ASYNC_CALLOUT_NOHUP))) {
			printk("check_modem_status: doing serial hangup...");
			if (info->tty)
				tty_hangup(info->tty);
		}
	}

	if ( (wake_counter != -1) && (counter > (wake_counter+100)) )
	{
		printk("about to wake %x\n", (unsigned int) info);
		wake_up_interruptible(&info->open_wait);
		wake_counter = -1;
	}

}

/*
 * This is the serial driver's interrupt routine for a single port
 */

/* Shay: parameters to this routine are always ( 0, NULL, NULL ) in this implementation */
/* Shay: activated from rs_timer to emulate interrupt */
static void rs_interrupt_single(int irq, void *dev_id, struct pt_regs * regs)
{
	int status;
/*  	int pass_counter = 0; */
	struct async_struct * info;
	struct async_struct * temp;
	int count = 0;
	static int cnt = 0;

	info = IRQ_ports[irq];
	temp = info;

/*  	printk("rs_interrupt_single (%d)\n", ++cnt); */

	if (!info || !info->tty)
		return;
	while ( temp->next_port != NULL && count < 4 )
	{
		count++;
		temp = temp->next_port;
	}
/*  	if ( count > 0 ) */
/*  	{ */
/*  		printk("rs_interrupt_single MORE THAN ONE INFO\n"); */
/*  	} */
	/* Shay: should find a way to get the status from the COM_IF so that we can
			 call transmit or receice only if necessary. meanwhile call them anyway */

	receive_chars(info, &status);
	check_modem_status(info);
	transmit_chars( info );

	info->last_active = jiffies;

/*    	printk("end.\n"); */
/*  #ifdef SERIAL_DEBUG_INTR */
/*  	printk("end.\n"); */
/*  #endif */
}

/*
 * -------------------------------------------------------------------
 * Here ends the serial interrupt routines.
 * -------------------------------------------------------------------
 */

/*
 * This routine is used to handle the "bottom half" processing for the
 * serial driver, known also the "software interrupt" processing.
 * This processing is done at the kernel interrupt level, after the
 * rs_interrupt() has returned, BUT WITH INTERRUPTS TURNED ON.  This
 * is where time-consuming activities which can not be done in the
 * interrupt driver proper are done; the interrupt driver schedules
 * them using rs_sched_event(), and they get done here.
 */
static void do_serial_bh(void)
{
//	//printk( "cnxt serial: do_serial_bh\n" );

	run_task_queue(&tq_serial);
}

static void do_softint(void *private_)
{
	struct async_struct	*info = (struct async_struct *) private_;
	struct tty_struct	*tty;

//	//printk( "cnxt serial: do_softint\n" );

	tty = info->tty;
	if (!tty)
		return;

	if (test_and_clear_bit(RS_EVENT_WRITE_WAKEUP, &info->event)) {
		if ((tty->flags & (1 << TTY_DO_WRITE_WAKEUP)) &&
		    tty->ldisc.write_wakeup)
			(tty->ldisc.write_wakeup)(tty);
		wake_up_interruptible(&tty->write_wait);
	}
}

/*
 * This subroutine is called when the RS_TIMER goes off.  It is used
 * by the serial driver to handle ports that do not have an interrupt
 * (irq=0).  This doesn't work very well for 16450's, but gives barely
 * passable results for a 16550A.  (Although at the expense of much
 * CPU overhead).
 */
/* dummy parameter to fit the timer_list prototype */
static void rs_timer( unsigned long unused )
{
/*  	printk("rs_timer at %ld\n", jiffies); */
	mod_timer(&serial_timer, jiffies + RS_STROBE_TIME);
	cnxt_task_struct.routine = cnxt_task;
	cnxt_task_struct.data = NULL;

#if ( LINUX_VERSION_CODE < KERNEL_VERSION( 2, 4, 0 ) )
	queue_task(&cnxt_task_struct, &tq_scheduler);
#else
	schedule_task(&cnxt_task_struct);
#endif

//	run_task_queue(&tq_cnxt);

/*  	printk("rs_timer modify to %ld\n", IRQ_timeout[0]); */

	mod_timer(&serial_timer, jiffies + IRQ_timeout[0]);
}

/*
 * ---------------------------------------------------------------
 * Low level utility subroutines for the serial driver:  routines to
 * figure out the appropriate timeout for an interrupt chain, routines
 * to initialize and startup a serial port, and routines to shutdown a
 * serial port.  Useful stuff like that.
 * ---------------------------------------------------------------
 */

/*
 * This routine figures out the correct timeout for a particular IRQ.
 * It uses the smallest timeout of all of the serial ports in a
 * particular interrupt chain.  Now only used for IRQ 0....
 */
static void figure_IRQ_timeout(int irq)
{
	struct	async_struct	*info;
	int	timeout = 60*HZ;	/* 60 seconds === a long time :-) */

	printk( "cnxt serial: figure_IRQ_timeout, irq=%d\n", irq );
	info = IRQ_ports[irq];
	if (!info) {
		IRQ_timeout[irq] = 60*HZ;
		printk( "figure_IRQ_timeout, IRQ_timeout[%d]=%d\n", irq,IRQ_timeout[irq] );
		return;
	}
	while (info) {
		if (info->timeout < timeout)
			timeout = info->timeout;
		info = info->next_port;
	}
	if (!irq)
		timeout = timeout / 2;
	IRQ_timeout[irq] = timeout ? timeout : 1;
	printk( "figure_IRQ_timeout, IRQ_timeout[%d]=%d\n", irq,IRQ_timeout[irq] );
}

static int startup(struct async_struct * info)
{
	unsigned long flags;
	int	retval=0;
	void (*handler)(int, void *, struct pt_regs *);
	struct serial_state *state= info->state;
	unsigned long page;

	//printk( "cnxt serial: startup\n" );

//	return 0; // modem ready

	page = get_free_page(GFP_KERNEL);
	if (!page)
	{
		printk("Page allocation failed !!! %x\n", (uint) info);
		return -ENOMEM;
	}

	save_flags(flags); cli();

	if (info->flags & ASYNC_INITIALIZED) {
		free_page(page);
		printk( "cnxt serial: startup error, info->flags & ASYNC_INITIALIZED %x\n", (uint) info );
		goto errout;
	}

	if (info->XMIT_BUF)
		free_page(page);
	else
		info->XMIT_BUF = (unsigned char *) page;

//	return 0; // modem did not respond

#ifdef SERIAL_DEBUG_OPEN
	printk("starting up ttys%d (irq %d)...", info->line, state->irq);
#endif

	/* Wake up UART and clear UART FIFOs (Shay: I don't think so) */
	serial_out( info, UART_MSR,
				serial_in(info, UART_MSR) | UART_MSR_DSR | UART_MSR_CTS );

#if 1
	 /* Allocate the IRQ if necessary (Shay: I don't think so) */
	handler = rs_interrupt_single;

#else /* Yair: Oh, really? */

	/*
	 * Allocate the IRQ if necessary
	 */
	if (state->irq && (!IRQ_ports[state->irq] ||
			  !IRQ_ports[state->irq]->next_port)) {
		if (IRQ_ports[state->irq]) {
#ifdef CONFIG_SERIAL_SHARE_IRQ
			free_irq(state->irq, NULL);
#ifdef CONFIG_SERIAL_MULTIPORT
			if (rs_multiport[state->irq].port1)
				handler = rs_interrupt_multi;
			else
#endif
				handler = rs_interrupt;
#else
			retval = -EBUSY;
			goto errout;
#endif /* CONFIG_SERIAL_SHARE_IRQ */
		} else
			handler = rs_interrupt_single;

		retval = request_irq(state->irq, handler, IRQ_T(info),
				     "serial", NULL);
		if (retval) {
			if (capable(CAP_SYS_ADMIN)) {
				if (info->tty)
					set_bit(TTY_IO_ERROR,
						&info->tty->flags);
				retval = 0;
			}
			goto errout;
		}
	}
#endif

	/*
	 * Insert serial port into IRQ chain.
	 */
	info->prev_port = 0;
	info->next_port = IRQ_ports[state->irq];
#ifdef SERIAL_DEBUG_OPEN
	printk("startup info->next_port = %x\n", (uint) info->next_port);
#endif
	if (info->next_port)
		info->next_port->prev_port = info;
#ifdef SERIAL_DEBUG_OPEN
	printk("startup IRQ_ports[%d] = %x\n", state->irq, (uint) info);
#endif
	IRQ_ports[state->irq] = info;
	figure_IRQ_timeout(state->irq);

	 /* Clear the interrupt registers. (Shay: I don't think so) */

	 /* Now, initialize the UART (Shay: I don't think so) */

	info->MCR = 0;
	if (info->tty->termios->c_cflag & CBAUD)
		info->MCR = UART_MCR_DTR | UART_MCR_RTS;
	{
		if (state->irq != 0)
			info->MCR |= UART_MCR_OUT2;
	}

	/*
	 * Finally, enable interrupts
	 */
	info->IER = UART_IER_MSI | UART_IER_RLSI | UART_IER_RDI;
//	serial_out(info, UART_IER, info->IER);	/* enable interrupts */

	 /* And clear the interrupt registers again for luck. (Shay: I don't think so) */

	if (info->tty)
		clear_bit(TTY_IO_ERROR, &info->tty->flags);
	info->XMIT_HEAD = info->XMIT_TAIL = 0;

	/*
	 * Set up serial timers...
	 */
	mod_timer(&serial_timer, jiffies + 2*HZ/100);

	/*
	 * Set up the tty->alt_speed kludge
	 */
	if (info->tty) {
		if ((info->flags & ASYNC_SPD_MASK) == ASYNC_SPD_HI)
			info->tty->alt_speed = 57600;
		if ((info->flags & ASYNC_SPD_MASK) == ASYNC_SPD_VHI)
			info->tty->alt_speed = 115200;
		if ((info->flags & ASYNC_SPD_MASK) == ASYNC_SPD_SHI)
			info->tty->alt_speed = 230400;
		if ((info->flags & ASYNC_SPD_MASK) == ASYNC_SPD_WARP)
			info->tty->alt_speed = 460800;
	}

	/*
	 * and set the speed of the serial port
	 */
	change_speed(info, 0);

	info->flags |= ASYNC_INITIALIZED;
	restore_flags(flags);
	return 0;

errout:
	//printk( "cnxt serial: startup returning error %d\n", retval );
	restore_flags(flags);
	return retval;
}

/*
 * This routine will shutdown a serial port; interrupts are disabled, and
 * DTR is dropped if the hangup on close termio flag is on.
 */
static void shutdown(struct async_struct * info)
{
	unsigned long	flags;
	struct serial_state *state;
//	int		retval;

	printk( "cnxt serial: shutdown\n" );

	if (!(info->flags & ASYNC_INITIALIZED))
		return;

	state = info->state;

#ifdef SERIAL_DEBUG_OPEN
	printk( "cnxt serial: shutdown, actually shuting down\n" );
	printk( "Shutting down serial port %d (irq %d)....",
			info->line, state->irq);
#endif

	save_flags(flags); cli(); /* Disable interrupts */

	/*
	 * clear delta_msr_wait queue to avoid mem leaks: we may free the irq
	 * here so the queue might never be waken up
	 */
#if ( LINUX_VERSION_CODE < KERNEL_VERSION( 2, 4, 0 ) )
	wake_up_interruptible(&info->delta_msr_wait);
#endif
	/*
	 * First unlink the serial port from the IRQ chain...
	 */
	if (info->next_port)
		info->next_port->prev_port = info->prev_port;
	if (info->prev_port)
		info->prev_port->next_port = info->next_port;
	else
	{
#ifdef SERIAL_DEBUG_OPEN
		printk("shutdown IRQ_ports[%d] = %x\n", state->irq, (uint) info->next_port);
#endif
		IRQ_ports[state->irq] = info->next_port;
	}
	figure_IRQ_timeout(state->irq);

	/*
	 * Free the IRQ, if necessary
	 */
//	if (state->irq && (!IRQ_ports[state->irq] ||
//			  !IRQ_ports[state->irq]->next_port)) {
//		if (IRQ_ports[state->irq]) {
//			free_irq(state->irq, NULL);
//			retval = request_irq(state->irq, rs_interrupt_single,
//					     IRQ_T(info), "serial", NULL);

//			if (retval)
//				//printk("serial shutdown: request_irq: error %d"
//				       "  Couldn't reacquire IRQ.\n", retval);
//		} else
//			free_irq(state->irq, NULL);
//	}

	if (info->XMIT_BUF) {
		free_page((unsigned long) info->XMIT_BUF);
		info->XMIT_BUF = 0;
	}

	info->IER = 0;
//	serial_out(info, UART_IER, 0x00);	/* disable all intrs */

	info->MCR &= ~UART_MCR_OUT2;

	/* disable break condition */
//	serial_out(info, UART_LCR, serial_in(info, UART_LCR) & ~UART_LCR_SBC);

	if (!info->tty || (info->tty->termios->c_cflag & HUPCL))
	{
#ifdef SERIAL_DEBUG_OPEN
		printk( "shutdown: lowering DTR & RTS\n" );
#endif
		info->MCR &= ~(UART_MCR_DTR|UART_MCR_RTS);
//		serial_out(info, UART_MCR, info->MCR);
	}

	/* disable FIFO's */
//	serial_out(info, UART_FCR, (UART_FCR_CLEAR_RCVR |
//				     UART_FCR_CLEAR_XMIT));
//	(void)serial_in(info, UART_RX);    /* read data port to reset things */

	if (info->tty)
	{
#ifdef SERIAL_DEBUG_OPEN
		printk( "shutdown: setting IO_ERROR bit in tty\n" );
#endif
		set_bit(TTY_IO_ERROR, &info->tty->flags);
	}

//	if (uart_config[info->state->type].flags & UART_STARTECH) {
		/* Arrange to enter sleep mode */
//		serial_out(info, UART_LCR, 0xBF);
//		serial_out(info, UART_EFR, UART_EFR_ECB);
//		serial_out(info, UART_IER, UART_IERX_SLEEP);
//		serial_out(info, UART_LCR, 0);
//	}
	if (info->state->type == PORT_16750) {
		/* Arrange to enter sleep mode */
//		serial_out(info, UART_IER, UART_IERX_SLEEP);
	}
	info->flags &= ~ASYNC_INITIALIZED;
	restore_flags(flags);
	printk( "shutdown: END\n" );
}

/*
 * This routine is called to set the UART divisor registers to match
 * the specified baud rate for a serial port.
 */
static void change_speed(struct async_struct *info,
			 struct termios *old_termios)
{
	unsigned short port;
	int	quot = 0, baud_base, baud;
	unsigned cflag, cval, fcr = 0;
	int	bits;
//	unsigned long	flags;

//	//printk( "cnxt serial: change_speed (returning immediatly)\n" );

	return; /* return immediatly for now */

	if (!info->tty || !info->tty->termios)
		return;
	cflag = info->tty->termios->c_cflag;
	if (!(port = info->port))
		return;

	/* byte size and parity */
	switch (cflag & CSIZE) {
	      case CS5: cval = 0x00; bits = 7; break;
	      case CS6: cval = 0x01; bits = 8; break;
	      case CS7: cval = 0x02; bits = 9; break;
	      case CS8: cval = 0x03; bits = 10; break;
	      /* Never happens, but GCC is too dumb to figure it out */
	      default:  cval = 0x00; bits = 7; break;
	      }
	if (cflag & CSTOPB) {
		cval |= 0x04;
		bits++;
	}
	if (cflag & PARENB) {
		cval |= UART_LCR_PARITY;
		bits++;
	}
	if (!(cflag & PARODD))
		cval |= UART_LCR_EPAR;
#ifdef CMSPAR
	if (cflag & CMSPAR)
		cval |= UART_LCR_SPAR;
#endif

	/* Determine divisor based on baud rate */
	baud = tty_get_baud_rate(info->tty);
	if (!baud)
		baud = 9600;	/* B0 transition handled in rs_set_termios */
	baud_base = info->state->baud_base;
	if (baud == 38400 &&
	    ((info->flags & ASYNC_SPD_MASK) == ASYNC_SPD_CUST))
		quot = info->state->custom_divisor;
	else {
		if (baud == 134)
			/* Special case since 134 is really 134.5 */
			quot = (2*baud_base / 269);
		else if (baud)
			quot = baud_base / baud;
	}
	/* If the quotient is zero refuse the change */
	if (!quot && old_termios) {
		info->tty->termios->c_cflag &= ~CBAUD;
		info->tty->termios->c_cflag |= (old_termios->c_cflag & CBAUD);
		baud = tty_get_baud_rate(info->tty);
		if (!baud)
			baud = 9600;
		if (baud == 38400 &&
		    ((info->flags & ASYNC_SPD_MASK) == ASYNC_SPD_CUST))
			quot = info->state->custom_divisor;
		else {
			if (baud == 134)
				/* Special case since 134 is really 134.5 */
				quot = (2*baud_base / 269);
			else if (baud)
				quot = baud_base / baud;
		}
	}
	/* As a last resort, if the quotient is zero, default to 9600 bps */
	if (!quot)
		quot = baud_base / 9600;
	info->quot = quot;
	info->timeout = ((info->xmit_fifo_size*HZ*bits*quot) / baud_base);
	info->timeout += HZ/50;		/* Add .02 seconds of slop */

	/* Set up FIFO's */
	if (uart_config[info->state->type].flags & UART_USE_FIFO) {
		if ((info->state->baud_base / quot) < 2400)
			fcr = UART_FCR_ENABLE_FIFO | UART_FCR_TRIGGER_1;
		else
			fcr = UART_FCR_ENABLE_FIFO | UART_FCR_TRIGGER_8;
	}
	if (info->state->type == PORT_16750)
		fcr |= UART_FCR7_64BYTE;

	/* CTS flow control flag and modem status interrupts */
	info->IER &= ~UART_IER_MSI;
	if (info->flags & ASYNC_HARDPPS_CD)
		info->IER |= UART_IER_MSI;
	if (cflag & CRTSCTS) {
		info->flags |= ASYNC_CTS_FLOW;
		info->IER |= UART_IER_MSI;
	} else
		info->flags &= ~ASYNC_CTS_FLOW;
	if (cflag & CLOCAL)
		info->flags &= ~ASYNC_CHECK_CD;
	else {
		info->flags |= ASYNC_CHECK_CD;
		info->IER |= UART_IER_MSI;
	}
	//serial_out(info, UART_IER, info->IER);

	/*
	 * Set up parity check flag
	 */
#define RELEVANT_IFLAG(iflag) (iflag & (IGNBRK|BRKINT|IGNPAR|PARMRK|INPCK))

	info->read_status_mask = UART_LSR_OE | UART_LSR_THRE | UART_LSR_DR;
	if (I_INPCK(info->tty))
		info->read_status_mask |= UART_LSR_FE | UART_LSR_PE;
	if (I_BRKINT(info->tty) || I_PARMRK(info->tty))
		info->read_status_mask |= UART_LSR_BI;

	/*
	 * Characters to ignore
	 */
	info->ignore_status_mask = 0;
	if (I_IGNPAR(info->tty))
		info->ignore_status_mask |= UART_LSR_PE | UART_LSR_FE;
	if (I_IGNBRK(info->tty)) {
		info->ignore_status_mask |= UART_LSR_BI;
		/*
		 * If we're ignore parity and break indicators, ignore
		 * overruns too.  (For real raw support).
		 */
		if (I_IGNPAR(info->tty))
			info->ignore_status_mask |= UART_LSR_OE;
	}
	/*
	 * !!! ignore all characters if CREAD is not set
	 */
	/*
	if ((cflag & CREAD) == 0)
		info->ignore_status_mask |= UART_LSR_DR;
	save_flags(flags); cli();
	if (uart_config[info->state->type].flags & UART_STARTECH) {
		//serial_out(info, UART_LCR, 0xBF);
		//serial_out(info, UART_EFR, (cflag & CRTSCTS) ? UART_EFR_CTS : 0);
	}
	*/
	/*
	serial_out(info, UART_LCR, cval | UART_LCR_DLAB);
	serial_out(info, UART_DLL, quot & 0xff);
	serial_out(info, UART_DLM, quot >> 8);
	if (info->state->type == PORT_16750)
		serial_out(info, UART_FCR, fcr);
	serial_out(info, UART_LCR, cval);
	if (info->state->type != PORT_16750)
		serial_out(info, UART_FCR, fcr);
	restore_flags(flags);
	*/
}

static void rs_put_char(struct tty_struct *tty, unsigned char ch)
{
	struct async_struct *info = (struct async_struct *)tty->driver_data;
	unsigned long flags;

/*  	printk( "cnxt serial: rs_put_char char=%c\n", ch ); */

	if (serial_paranoia_check(info, tty->device, "rs_put_char"))
		return;

	if (!tty || !info->XMIT_BUF)
		return;

	save_flags(flags); cli();
	if ( CIRC_SPACE(info->XMIT_HEAD, info->XMIT_TAIL, SERIAL_XMIT_SIZE) == 0 )
	{
		restore_flags(flags);
		return;
	}

	info->XMIT_BUF[info->XMIT_HEAD++] = ch;
	info->XMIT_HEAD &= SERIAL_XMIT_SIZE-1;

	restore_flags(flags);
}

static void rs_flush_chars(struct tty_struct *tty)
{
	struct async_struct *info = (struct async_struct *)tty->driver_data;
	/*	unsigned long flags; */

//	//printk( "cnxt serial: rs_flush_chars\n" );

	if (serial_paranoia_check(info, tty->device, "rs_flush_chars"))
		return;

	if (info->XMIT_HEAD == info->XMIT_TAIL || tty->stopped || tty->hw_stopped ||
	    !info->XMIT_BUF)
		return;
	/*
	save_flags(flags); cli();
	*/
	info->IER |= UART_IER_THRI;
	/*
	serial_out(info, UART_IER, info->IER);
	restore_flags(flags);
	*/
}

static int rs_write(struct tty_struct* tty, int from_user, const unsigned char* buf, int count)
{
	int	c, ret = 0;
	struct async_struct *info = (struct async_struct *)tty->driver_data;
	unsigned long flags;

        //printk( "cnxt serial: rs_write %d bytes, %s\n", count, buf );

	if (serial_paranoia_check(info, tty->device, "rs_write"))
		return 0;

	if (!tty || !info->XMIT_BUF || !tmp_buf) {
		//printk( "cnxt serial: rs_write tty=%x, info->xmit_buf=%x, tmp_buf=%x\n", tty, info->xmit_buf, tmp_buf );
		return 0;
	}

	save_flags(flags);
	if (from_user) {
		down(&tmp_buf_sem);
		while (1) {
			int c1;

			c = CIRC_SPACE_TO_END(info->XMIT_HEAD,
                                  info->XMIT_TAIL,
                                  SERIAL_XMIT_SIZE);

            if (count < c)
				c = count;
			if (c <= 0)
				break;

			c -= copy_from_user(tmp_buf, buf, c);
			if (!c) {
				if (!ret)
					ret = -EFAULT;
				break;
			}
			cli();
            c1 = CIRC_SPACE_TO_END(info->XMIT_HEAD,
                                   info->XMIT_TAIL,
                                   SERIAL_XMIT_SIZE);
            if (c1 < c)
				c = c1;

			memcpy(info->XMIT_BUF + info->XMIT_HEAD, tmp_buf, c);
			info->XMIT_HEAD = ((info->XMIT_HEAD + c) &
					   (SERIAL_XMIT_SIZE-1));
			restore_flags(flags);
			buf += c;
			count -= c;
			ret += c;
		}
		up(&tmp_buf_sem);
	} else {
		cli();
		while (1) {
			c = CIRC_SPACE_TO_END(info->XMIT_HEAD,
                                  info->XMIT_TAIL,
                                  SERIAL_XMIT_SIZE);
            if (count < c)
				c = count;
            if (c <= 0)
                break;

			memcpy(info->XMIT_BUF + info->XMIT_HEAD, buf, c);
			info->XMIT_HEAD = ((info->XMIT_HEAD + c) &
					   (SERIAL_XMIT_SIZE-1));
			buf += c;
			count -= c;
			ret += c;
		}
		restore_flags(flags);
	}
	if ( info->XMIT_HEAD != info->XMIT_TAIL && 
         !tty->stopped && 
         !tty->hw_stopped &&
	     !(info->IER & UART_IER_THRI)) {
		info->IER |= UART_IER_THRI;
	}

//	if ( count != 0 )
//		printk( "rs_write: got only %d out of %d\n", ret, count+ret );

//    printk( "cnxt serial: rs_write returning %d\n", ret );
	return ret;
}

static int rs_write_room(struct tty_struct *tty)
{
	struct async_struct *info = (struct async_struct *)tty->driver_data;
	int	ret;

//	//printk( "cnxt serial: rs_write_room\n" );

	if (serial_paranoia_check(info, tty->device, "rs_write_room"))
		return 0;
	ret = CIRC_SPACE(info->XMIT_HEAD, info->XMIT_TAIL, SERIAL_XMIT_SIZE);;
	if (ret < 0)
		ret = 0;
	return ret;
}

static int rs_chars_in_buffer(struct tty_struct *tty)
{
	struct async_struct *info = (struct async_struct *)tty->driver_data;

//	printk( "cnxt serial: rs_chars_in_buffer\n" );

	if (serial_paranoia_check(info, tty->device, "rs_chars_in_buffer"))
		return 0;

/*	//printk( "cnxt serial: rs_chars_in_buffer returning %d\n", info->xmit_cnt ); */

	return ((info->XMIT_HEAD - info->XMIT_TAIL) & (SERIAL_XMIT_SIZE-1));
}

static void rs_flush_buffer(struct tty_struct *tty)
{
	struct async_struct *info = (struct async_struct *)tty->driver_data;
	unsigned long flags;

//	//printk( "cnxt serial: rs_flush_buffer\n" );

	if (serial_paranoia_check(info, tty->device, "rs_flush_buffer"))
		return;
	save_flags(flags); cli();
	info->XMIT_HEAD = info->XMIT_TAIL = 0;
	restore_flags(flags);
	wake_up_interruptible(&tty->write_wait);
	if ((tty->flags & (1 << TTY_DO_WRITE_WAKEUP)) &&
	    tty->ldisc.write_wakeup)
		(tty->ldisc.write_wakeup)(tty);
}

/*
 * This function is used to send a high-priority XON/XOFF character to
 * the device
 */
static void rs_send_xchar(struct tty_struct *tty, char ch)
{
	struct async_struct *info = (struct async_struct *)tty->driver_data;

//	//printk( "cnxt serial: rs_send_xchar\n" );

	if (serial_paranoia_check(info, tty->device, "rs_send_char"))
		return;

	info->x_char = ch;
	if (ch) {
	/* Make sure transmit interrupts are on */
	    info->IER |= UART_IER_THRI;
	    //serial_out(info, UART_IER, info->IER);
	}
}

/*
 * ------------------------------------------------------------
 * rs_throttle()
 *
 * This routine is called by the upper-layer tty layer to signal that
 * incoming characters should be throttled.
 * ------------------------------------------------------------
 */
static void rs_throttle(struct tty_struct * tty)
{
	struct async_struct *info = (struct async_struct *)tty->driver_data;
	/*	unsigned long flags; */

	//printk( "cnxt serial: rs_throttle\n" );

#ifdef SERIAL_DEBUG_THROTTLE
	char	buf[64];

	//printk("throttle %s: %d....\n", tty_name(tty, buf),
	//       tty->ldisc.chars_in_buffer(tty));
#endif

	if (serial_paranoia_check(info, tty->device, "rs_throttle"))
		return;

	if (I_IXOFF(tty))
		rs_send_xchar(tty, STOP_CHAR(tty));

	if (tty->termios->c_cflag & CRTSCTS)
		info->MCR &= ~UART_MCR_RTS;
	/*
	save_flags(flags); cli();
	serial_out(info, UART_MCR, info->MCR);
	restore_flags(flags);
	*/
}

static void rs_unthrottle(struct tty_struct * tty)
{
	struct async_struct *info = (struct async_struct *)tty->driver_data;
//	unsigned long flags;

	//printk( "cnxt serial: rs_unthrottle\n" );

#ifdef SERIAL_DEBUG_THROTTLE
	char	buf[64];

	//printk("unthrottle %s: %d....\n", tty_name(tty, buf),
	 //      tty->ldisc.chars_in_buffer(tty));
#endif

	if (serial_paranoia_check(info, tty->device, "rs_unthrottle"))
		return;

	if (I_IXOFF(tty))
	{
	    if (info->x_char)
		info->x_char = 0;
	    else
		rs_send_xchar(tty, START_CHAR(tty));
	}
	if (tty->termios->c_cflag & CRTSCTS)
		info->MCR |= UART_MCR_RTS;
	/*
	save_flags(flags); cli();
	serial_out(info, UART_MCR, info->MCR);
	restore_flags(flags);
	*/
}

/*
 * ------------------------------------------------------------
 * rs_ioctl() and friends
 * ------------------------------------------------------------
 */

static int get_serial_info(struct async_struct * info,
			   struct serial_struct * retinfo)
{
	struct serial_struct tmp;
	struct serial_state *state = info->state;

	//printk( "cnxt serial: get_serial_info\n" );

	if (!retinfo)
		return -EFAULT;
	memset(&tmp, 0, sizeof(tmp));
	tmp.type = state->type;
	tmp.line = state->line;
	tmp.port = state->port;
	tmp.irq = state->irq;
	tmp.flags = state->flags;
	tmp.xmit_fifo_size = state->xmit_fifo_size;
	tmp.baud_base = state->baud_base;
	tmp.close_delay = state->close_delay;
	tmp.closing_wait = state->closing_wait;
	tmp.custom_divisor = state->custom_divisor;
	tmp.hub6 = state->hub6;
	if (copy_to_user(retinfo,&tmp,sizeof(*retinfo)))
		return -EFAULT;
	return 0;
}

static int set_serial_info(struct async_struct * info,
			   struct serial_struct * new_info)
{
	struct serial_struct new_serial;
 	struct serial_state old_state, *state;
	unsigned int		i,change_irq,change_port;
	int 			retval = 0;

	//printk( "cnxt serial: set_serial_info\n" );

	if (copy_from_user(&new_serial,new_info,sizeof(new_serial)))
		return -EFAULT;
	state = info->state;
	old_state = *state;

	change_irq = new_serial.irq != state->irq;
	change_port = (new_serial.port != state->port) ||
		(new_serial.hub6 != state->hub6);

	if (!capable(CAP_SYS_ADMIN)) {
		if (change_irq || change_port ||
		    (new_serial.baud_base != state->baud_base) ||
		    (new_serial.type != state->type) ||
		    (new_serial.close_delay != state->close_delay) ||
		    (new_serial.xmit_fifo_size != state->xmit_fifo_size) ||
		    ((new_serial.flags & ~ASYNC_USR_MASK) !=
		     (state->flags & ~ASYNC_USR_MASK)))
			return -EPERM;
		state->flags = ((state->flags & ~ASYNC_USR_MASK) |
			       (new_serial.flags & ASYNC_USR_MASK));
		info->flags = ((info->flags & ~ASYNC_USR_MASK) |
			       (new_serial.flags & ASYNC_USR_MASK));
		state->custom_divisor = new_serial.custom_divisor;
		goto check_and_exit;
	}

	new_serial.irq = irq_cannonicalize(new_serial.irq);

	if ((new_serial.irq >= NR_IRQS) || (new_serial.port > 0xffff) ||
	    (new_serial.baud_base < 9600)|| (new_serial.type < PORT_UNKNOWN) ||
	    (new_serial.type > PORT_MAX) || (new_serial.type == PORT_CIRRUS) ||
	    (new_serial.type == PORT_STARTECH)) {
		return -EINVAL;
	}

	if ((new_serial.type != state->type) ||
	    (new_serial.xmit_fifo_size <= 0))
		new_serial.xmit_fifo_size =
			uart_config[state->type].dfl_xmit_fifo_size;

	/* Make sure address is not already in use */
	if (new_serial.type) {
		for (i = 0 ; i < NR_PORTS; i++)
			if ((state != &rs_table[i]) &&
			    (rs_table[i].port == new_serial.port) &&
			    rs_table[i].type)
				return -EADDRINUSE;
	}

	if ((change_port || change_irq) && (state->count > 1))
		return -EBUSY;

	/*
	 * OK, past this point, all the error checking has been done.
	 * At this point, we start making changes.....
	 */

	state->baud_base = new_serial.baud_base;
	state->flags = ((state->flags & ~ASYNC_FLAGS) |
			(new_serial.flags & ASYNC_FLAGS));
	info->flags = ((state->flags & ~ASYNC_INTERNAL_FLAGS) |
		       (info->flags & ASYNC_INTERNAL_FLAGS));
	state->custom_divisor = new_serial.custom_divisor;
	state->type = new_serial.type;
	state->close_delay = new_serial.close_delay * HZ/100;
	state->closing_wait = new_serial.closing_wait * HZ/100;
	info->tty->low_latency = (info->flags & ASYNC_LOW_LATENCY) ? 1 : 0;
	info->xmit_fifo_size = state->xmit_fifo_size =
		new_serial.xmit_fifo_size;

	release_region(state->port,8);
	if (change_port || change_irq) {
		/*
		 * We need to shutdown the serial port at the old
		 * port/irq combination.
		 */
		printk( "set_serial_info calling shutdown...\n") ;
		shutdown(info);
		state->irq = new_serial.irq;
		info->port = state->port = new_serial.port;
		info->hub6 = state->hub6 = new_serial.hub6;
	}
	if (state->type != PORT_UNKNOWN)
		request_region(state->port,8,"serial(set)");


check_and_exit:
	if (!state->port || !state->type)
		return 0;
	if (info->flags & ASYNC_INITIALIZED) {
		if (((old_state.flags & ASYNC_SPD_MASK) !=
		     (state->flags & ASYNC_SPD_MASK)) ||
		    (old_state.custom_divisor != state->custom_divisor)) {
			if ((state->flags & ASYNC_SPD_MASK) == ASYNC_SPD_HI)
				info->tty->alt_speed = 57600;
			if ((state->flags & ASYNC_SPD_MASK) == ASYNC_SPD_VHI)
				info->tty->alt_speed = 115200;
			if ((state->flags & ASYNC_SPD_MASK) == ASYNC_SPD_SHI)
				info->tty->alt_speed = 230400;
			if ((state->flags & ASYNC_SPD_MASK) == ASYNC_SPD_WARP)
				info->tty->alt_speed = 460800;
			change_speed(info, 0);
		}
	} else
	{
		//printk("set_serial_info calling startup...\n");
		retval = startup(info);
	}
	return retval;
}


/*
 * get_lsr_info - get line status register info
 *
 * Purpose: Let user call ioctl() to get info when the UART physically
 * 	    is emptied.  On bus types like RS485, the transmitter must
 * 	    release the bus after transmitting. This must be done when
 * 	    the transmit shift register is empty, not be done when the
 * 	    transmit holding register is empty.  This functionality
 * 	    allows an RS485 driver to be written in user space.
 */
static int get_lsr_info(struct async_struct * info, unsigned int *value)
{
	unsigned char status;
	unsigned int result;
	unsigned long flags;

	//printk( "cnxt serial: get_lsr_info\n" );

	save_flags(flags); cli();
	status = serial_in(info, UART_LSR);
	restore_flags(flags);
	result = ((status & UART_LSR_TEMT) ? TIOCSER_TEMT : 0);
	return put_user(result,value);
}


static int get_modem_info(struct async_struct * info, unsigned int *value)
{
	unsigned char control, status;
	unsigned int result;

//	printk( "cnxt serial: get_modem_info\n" );

	control = info->MCR;
	status = serial_in( info, UART_MSR );

	result = ( ((control & UART_MCR_RTS) ? TIOCM_RTS : 0) |
			   ((control & UART_MCR_DTR) ? TIOCM_DTR : 0) |
			   ((status  & UART_MSR_DCD) ? TIOCM_CAR : 0) |
			   ((status  & UART_MSR_RI)  ? TIOCM_RNG : 0) |
			   ((status  & UART_MSR_DSR) ? TIOCM_DSR : 0) |
			   ((status  & UART_MSR_CTS) ? TIOCM_CTS : 0) );

#if 0
	printk( "  MCR/MSR: 0x%x(%s%s%s%s%s%s)\n",
			result,
			(result & TIOCM_RTS) ? "RTS|" : "",
			(result & TIOCM_DTR) ? "DTR|" : "",
			(result & TIOCM_CAR) ? "DCD|" : "",
			(result & TIOCM_RNG) ? "RNG|" : "",
			(result & TIOCM_DSR) ? "DSR|" : "",
			(result & TIOCM_CTS) ? "CTS" : "" );
#endif

	return put_user(result,value);
}

static int set_modem_info(struct async_struct * info, unsigned int cmd,
			  unsigned int *value)
{
	int error;
	unsigned int arg;
//	unsigned long flags;

//	//printk( "cnxt serial: set_modem_info\n" );

	error = get_user(arg, value);
	if (error)
		return error;
	switch (cmd) {
	case TIOCMBIS:
		if (arg & TIOCM_RTS)
			info->MCR |= UART_MCR_RTS;
		if (arg & TIOCM_DTR)
			info->MCR |= UART_MCR_DTR;
#ifdef TIOCM_OUT1
		if (arg & TIOCM_OUT1)
			info->MCR |= UART_MCR_OUT1;
		if (arg & TIOCM_OUT2)
			info->MCR |= UART_MCR_OUT2;
#endif
		break;
	case TIOCMBIC:
		if (arg & TIOCM_RTS)
			info->MCR &= ~UART_MCR_RTS;
		if (arg & TIOCM_DTR)
			info->MCR &= ~UART_MCR_DTR;
#ifdef TIOCM_OUT1
		if (arg & TIOCM_OUT1)
			info->MCR &= ~UART_MCR_OUT1;
		if (arg & TIOCM_OUT2)
			info->MCR &= ~UART_MCR_OUT2;
#endif
		break;
	case TIOCMSET:
		info->MCR = ((info->MCR & ~(UART_MCR_RTS |
#ifdef TIOCM_OUT1
					    UART_MCR_OUT1 |
					    UART_MCR_OUT2 |
#endif
					    UART_MCR_DTR))
			     | ((arg & TIOCM_RTS) ? UART_MCR_RTS : 0)
#ifdef TIOCM_OUT1
			     | ((arg & TIOCM_OUT1) ? UART_MCR_OUT1 : 0)
			     | ((arg & TIOCM_OUT2) ? UART_MCR_OUT2 : 0)
#endif
			     | ((arg & TIOCM_DTR) ? UART_MCR_DTR : 0));
		break;
	default:
		return -EINVAL;
	}
	/*
	save_flags(flags); cli();
	serial_out(info, UART_MCR, info->MCR);
	restore_flags(flags);
	*/
	return 0;
}

static int do_autoconfig(struct async_struct * info)
{
	int			retval;

	//printk( "cnxt serial: do_autoconfig\n" );

	if (!capable(CAP_SYS_ADMIN))
		return -EPERM;

	if (info->state->count > 1)
		return -EBUSY;

	printk( "do_autoconfig calling shutdown...\n") ;
	shutdown(info);

	autoconfig(info->state);
	if ((info->state->flags & ASYNC_AUTO_IRQ) &&
	    (info->state->port != 0) &&
	    (info->state->type != PORT_UNKNOWN))
		info->state->irq = 0; /* Shay: use 0 for now ?!? */

	printk("do_autoconfig calling startup...\n");
	retval = startup(info);
	if (retval)
		return retval;
	return 0;
}

/*
 * rs_break() --- routine which turns the break handling on or off
 */
static void rs_break(struct tty_struct *tty, int break_state)
{
	struct async_struct * info = (struct async_struct *)tty->driver_data;
//	unsigned long flags;

	//printk( "cnxt serial: rs_break\n" );

	return; /* quit here for now */

	if (serial_paranoia_check(info, tty->device, "rs_break"))
		return;

	if (!info->port)
		return;
	/*
	save_flags(flags); cli();
	if (break_state == -1)
		serial_out(info, UART_LCR, serial_in(info, UART_LCR) | UART_LCR_SBC);
	else
		serial_out(info, UART_LCR, serial_in(info, UART_LCR) & ~UART_LCR_SBC);
	restore_flags(flags);
	*/
}

static int rs_ioctl(struct tty_struct *tty, struct file * file,
		    unsigned int cmd, unsigned long arg)
{
	int error;
	struct async_struct * info = (struct async_struct *)tty->driver_data;
	struct async_icount cprev, cnow;	/* kernel counter temps */
	struct serial_icounter_struct *p_cuser;	/* user space */
	unsigned long flags;

//	printk( "cnxt serial: rs_ioctl cmd = %d\n", cmd );

	if (serial_paranoia_check(info, tty->device, "rs_ioctl"))
		return -ENODEV;

	if ((cmd != TIOCGSERIAL) && (cmd != TIOCSSERIAL) &&
	    (cmd != TIOCSERCONFIG) && (cmd != TIOCSERGSTRUCT) &&
	    (cmd != TIOCMIWAIT) && (cmd != TIOCGICOUNT)) {
		if (tty->flags & (1 << TTY_IO_ERROR))
		    return -EIO;
	}

	switch (cmd) {
		case TIOCMGET:
			//printk( "cnxt serial: rs_ioctl TIOCMGET\n" );
			return get_modem_info(info, (unsigned int *) arg);
		case TIOCMBIS:
		case TIOCMBIC:
		case TIOCMSET:
			//printk( "cnxt serial: rs_ioctl TIOCMBIS || TIOCMBIC || TIOCMSET\n" );
			return set_modem_info(info, cmd, (unsigned int *) arg);
		case TIOCGSERIAL:
			//printk( "cnxt serial: rs_ioctl TIOCGSERIAL\n" );
			return get_serial_info(info,
					       (struct serial_struct *) arg);
		case TIOCSSERIAL:
			//printk( "cnxt serial: rs_ioctl TIOCSSERIAL\n" );
			return set_serial_info(info,
					       (struct serial_struct *) arg);
		case TIOCSERCONFIG:
			//printk( "cnxt serial: rs_ioctl TIOCSERCONFIG\n" );
			return do_autoconfig(info);

		case TIOCSERGETLSR: /* Get line status register */
			//printk( "cnxt serial: rs_ioctl TIOCSERGETLSR\n" );
			return get_lsr_info(info, (unsigned int *) arg);

		case TIOCSERGSTRUCT:
			//printk( "cnxt serial: rs_ioctl TIOCSERGSTRUCT\n" );
			if (copy_to_user((struct async_struct *) arg,
					 info, sizeof(struct async_struct)))
				return -EFAULT;
			return 0;

		/* Shay: next 2 values are for multi port only and aren't supported */
		case TIOCSERGETMULTI:
			//printk( "cnxt serial: rs_ioctl TIOCSERGETMULTI\n" );
			return -EFAULT;

		case TIOCSERSETMULTI:
			//printk( "cnxt serial: rs_ioctl TIOCSERSETMULTI\n" );
			return -EFAULT;


		/*
		 * Wait for any of the 4 modem inputs (DCD,RI,DSR,CTS) to change
		 * - mask passed in arg for lines of interest
 		 *   (use |'ed TIOCM_RNG/DSR/CD/CTS for masking)
		 * Caller should use TIOCGICOUNT to see which one it was
		 */
		case TIOCMIWAIT:
			//printk( "cnxt serial: rs_ioctl TIOCMIWAIT\n" );
			save_flags(flags); cli();
			/* note the counters on entry */
			cprev = info->state->icount;
			restore_flags(flags);
			while (1) {
				interruptible_sleep_on(&info->delta_msr_wait);
				/* see if a signal did it */
				if (signal_pending(current))
					return -ERESTARTSYS;
				save_flags(flags); cli();
				cnow = info->state->icount; /* atomic copy */
				restore_flags(flags);
				if (cnow.rng == cprev.rng && cnow.dsr == cprev.dsr &&
				    cnow.dcd == cprev.dcd && cnow.cts == cprev.cts)
					return -EIO; /* no change => error */
				if ( ((arg & TIOCM_RNG) && (cnow.rng != cprev.rng)) ||
				     ((arg & TIOCM_DSR) && (cnow.dsr != cprev.dsr)) ||
				     ((arg & TIOCM_CD)  && (cnow.dcd != cprev.dcd)) ||
				     ((arg & TIOCM_CTS) && (cnow.cts != cprev.cts)) ) {
					return 0;
				}
				cprev = cnow;
			}
			/* NOTREACHED */

		/*
		 * Get counter of input serial line interrupts (DCD,RI,DSR,CTS)
		 * Return: write counters to the user passed counter struct
		 * NB: both 1->0 and 0->1 transitions are counted except for
		 *     RI where only 0->1 is counted.
		 */
		case TIOCGICOUNT:
			//printk( "cnxt serial: rs_ioctl TIOCGICOUNT\n" );
			save_flags(flags); cli();
			cnow = info->state->icount;
			restore_flags(flags);
			p_cuser = (struct serial_icounter_struct *) arg;
			error = put_user(cnow.cts, &p_cuser->cts);
			if (error) return error;
			error = put_user(cnow.dsr, &p_cuser->dsr);
			if (error) return error;
			error = put_user(cnow.rng, &p_cuser->rng);
			if (error) return error;
			error = put_user(cnow.dcd, &p_cuser->dcd);
			if (error) return error;
			error = put_user(cnow.rx, &p_cuser->rx);
			if (error) return error;
			error = put_user(cnow.tx, &p_cuser->tx);
			if (error) return error;
			error = put_user(cnow.frame, &p_cuser->frame);
			if (error) return error;
			error = put_user(cnow.overrun, &p_cuser->overrun);
			if (error) return error;
			error = put_user(cnow.parity, &p_cuser->parity);
			if (error) return error;
			error = put_user(cnow.brk, &p_cuser->brk);
			if (error) return error;
			error = put_user(cnow.buf_overrun, &p_cuser->buf_overrun);
			if (error) return error;
			return 0;

		case TIOCSERGWILD:
		case TIOCSERSWILD:
			//printk( "cnxt serial: rs_ioctl TIOCSERGWILD || TIOCSERSWILD\n" );
			/* "setserial -W" is called in Debian boot */
			//printk ("TIOCSER?WILD ioctl obsolete, ignored.\n");
			return 0;

		default:
			//printk( "cnxt serial: rs_ioctl default\n" );
			return -ENOIOCTLCMD;
		}
	return 0;
}

static void rs_set_termios(struct tty_struct *tty, struct termios *old_termios)
{
	struct async_struct *info = (struct async_struct *)tty->driver_data;
	/*unsigned long flags;*/
	unsigned int cflag = tty->termios->c_cflag;

//	return; /* Shay: quit for now */

	if (   (cflag == old_termios->c_cflag)
	    && (   RELEVANT_IFLAG(tty->termios->c_iflag)
		== RELEVANT_IFLAG(old_termios->c_iflag)))
	  return;

//	change_speed(info, old_termios);

	/* Handle transition to B0 status */
	if ((old_termios->c_cflag & CBAUD) && !(cflag & CBAUD))
	{
		info->MCR &= ~(UART_MCR_DTR|UART_MCR_RTS);
		//printk( "cnxt serial: rs_set_termios UART_MCR_DTR and UART_MCR_RTS off\n" );
		/*
		save_flags(flags); cli();
		serial_out(info, UART_MCR, info->MCR);
		restore_flags(flags);
		*/
	}

	/* Handle transition away from B0 status */
	if (!(old_termios->c_cflag & CBAUD) && (cflag & CBAUD))
	{
		info->MCR |= UART_MCR_DTR;
		if (!(tty->termios->c_cflag & CRTSCTS) || !test_bit(TTY_THROTTLED, &tty->flags))
		{
		    //printk( "cnxt serial: rs_set_termios UART_MCR_RTS on\n" );
		    info->MCR |= UART_MCR_RTS;
		}
		/*
		save_flags(flags); cli();
		serial_out(info, UART_MCR, info->MCR);
		restore_flags(flags);
		*/
	}

	return;
	/* Handle turning off CRTSCTS */
	if ((old_termios->c_cflag & CRTSCTS) &&
	    !(tty->termios->c_cflag & CRTSCTS)) {
		tty->hw_stopped = 0;
		rs_start(tty);
	}

}

/*
 * ------------------------------------------------------------
 * rs_close()
 *
 * This routine is called when the serial port gets closed.  First, we
 * wait for the last remaining data to be sent.  Then, we unlink its
 * async structure from the interrupt chain if necessary, and we free
 * that IRQ if nothing is left in the chain.
 * ------------------------------------------------------------
 */
static void rs_close(struct tty_struct *tty, struct file * filp)
{
	struct async_struct * info = (struct async_struct *)tty->driver_data;
	struct serial_state *state;
	unsigned long flags;

	printk( "cnxt serial: rs_close\n" );

	if (!info || serial_paranoia_check(info, tty->device, "rs_close"))
		return;

	printk( "cnxt serial: run all th etasts before closing...\n" );
	
#if ( LINUX_VERSION_CODE < KERNEL_VERSION( 2, 4, 0 ) )
	run_task_queue(&tq_serial);
	run_task_queue(&tq_scheduler);
#else
	while ( list_empty(&tq_serial) == FALSE )
	{
		printk( "cnxt derial:: list_del 0x%x.\n",tq_serial.next );
		list_del(tq_serial.next);
	}

	flush_scheduled_tasks();
#endif
	run_task_queue(&tq_immediate);

	state = info->state;

	save_flags(flags); cli();

	if (tty_hung_up_p(filp)) {
		printk("before DEC-hung\n");
		DBG_CNT("before DEC-hung");
		MOD_DEC_USE_COUNT;
		restore_flags(flags);
		return;
	}

	printk("rs_close /dev/cnxt%d, count=%d (tty->count=%d)\n",
		   info->line, state->count, tty->count );
#ifdef SERIAL_DEBUG_OPEN
	printk("rs_close ttys%d, count = %d\n", info->line, state->count);
#endif
	if ((tty->count == 1) && (state->count != 1)) {
		/*
		 * Uh, oh.  tty->count is 1, which means that the tty
		 * structure will be freed.  state->count should always
		 * be one in these conditions.  If it's greater than
		 * one, we've got real problems, since it means the
		 * serial port won't be shutdown.
		 */
		printk("rs_close: bad serial port count; tty->count is 1, "
		       "state->count is %d\n", state->count);
		state->count = 1;
	}
	if (--state->count < 0) {
		//printk("rs_close: bad serial port count for ttys%d: %d\n",
		 //      info->line, state->count);
		state->count = 0;
	}
	if (state->count) {
		printk("before DEC-2\n");
		DBG_CNT("before DEC-2");
		MOD_DEC_USE_COUNT;
		restore_flags(flags);
		return;
	}
	info->flags |= ASYNC_CLOSING;
	/*
	 * Save the termios structure, since this port may have
	 * separate termios for callout and dialin.
	 */
	if (info->flags & ASYNC_NORMAL_ACTIVE)
		info->state->normal_termios = *tty->termios;
	if (info->flags & ASYNC_CALLOUT_ACTIVE)
		info->state->callout_termios = *tty->termios;
	/*
	 * Now we wait for the transmit buffer to clear; and we notify
	 * the line discipline to only process XON/XOFF characters.
	 */
	tty->closing = 1;
	if (info->closing_wait != ASYNC_CLOSING_WAIT_NONE)
		tty_wait_until_sent(tty, info->closing_wait);
	/*
	 * At this point we stop accepting input.  To do this, we
	 * disable the receive line status interrupts, and tell the
	 * interrupt driver to stop checking the data ready bit in the
	 * line status register.
	 */
	/*
	info->IER &= ~UART_IER_RLSI;
	info->read_status_mask &= ~UART_LSR_DR;
	if (info->flags & ASYNC_INITIALIZED) {
		serial_out(info, UART_IER, info->IER);
		*/
		/*
		 * Before we drop DTR, make sure the UART transmitter
		 * has completely drained; this is especially
		 * important if there is a transmit FIFO!
		rs_wait_until_sent(tty, info->timeout);

	}
		 */
#ifdef SERIAL_DEBUG_OPEN
	printk( "rs_close: calling shutdown\n" );
#endif

	shutdown(info);
	if (tty->driver.flush_buffer)
		tty->driver.flush_buffer(tty);


	if (tty->ldisc.flush_buffer)
		tty->ldisc.flush_buffer(tty);
	tty->closing = 0;
	info->event = 0;
	info->tty = 0;
	if (info->blocked_open) {
		if (info->close_delay) {
			current->state = TASK_INTERRUPTIBLE;
			schedule_timeout(info->close_delay);
		}
		wake_up_interruptible(&info->open_wait);
	}
	info->flags &= ~(ASYNC_NORMAL_ACTIVE|ASYNC_CALLOUT_ACTIVE|
			 ASYNC_CLOSING);
/*  	wake_up_interruptible(&info->close_wait); */
	MOD_DEC_USE_COUNT;

	restore_flags(flags);
	printk( "rs_close: END\n" );
}

/*
 * rs_wait_until_sent() --- wait until the transmitter is empty
 */
static void rs_wait_until_sent(struct tty_struct *tty, int timeout)
{
	struct async_struct * info = (struct async_struct *)tty->driver_data;
	unsigned long orig_jiffies, char_time;
	int lsr;

	//printk( "cnxt serial: rs_wait_until_sent\n" );

	if (serial_paranoia_check(info, tty->device, "rs_wait_until_sent"))
		return;

	if (info->state->type == PORT_UNKNOWN)
		return;

	if (info->xmit_fifo_size == 0)
		return; /* Just in case.... */

	orig_jiffies = jiffies;
	/*
	 * Set the check interval to be 1/5 of the estimated time to
	 * send a single character, and make it at least 1.  The check
	 * interval should also be less than the timeout.
	 *
	 * Note: we have to use pretty tight timings here to satisfy
	 * the NIST-PCTS.
	 */
	char_time = (info->timeout - HZ/50) / info->xmit_fifo_size;
	char_time = char_time / 5;
	if (char_time == 0)
		char_time = 1;
	if (timeout)
	  char_time = MIN(char_time, timeout);
	/*
	 * If the transmitter hasn't cleared in twice the approximate
	 * amount of time to send the entire FIFO, it probably won't
	 * ever clear.  This assumes the UART isn't doing flow
	 * control, which is currently the case.  Hence, if it ever
	 * takes longer than info->timeout, this is probably due to a
	 * UART bug of some kind.  So, we clamp the timeout parameter at
	 * 2*info->timeout.
	 */
	if (!timeout || timeout > 2*info->timeout)
		timeout = 2*info->timeout;
#ifdef SERIAL_DEBUG_RS_WAIT_UNTIL_SENT
	//printk("In rs_wait_until_sent(%d) check=%lu...", timeout, char_time);
	//printk("jiff=%lu...", jiffies);
#endif
	while (!((lsr = serial_in(info, UART_LSR)) & UART_LSR_TEMT)) {
#ifdef SERIAL_DEBUG_RS_WAIT_UNTIL_SENT
		//printk("lsr = %d (jiff=%lu)...", lsr, jiffies);
#endif
		current->state = TASK_INTERRUPTIBLE;
		current->counter = 0;	/* make us low-priority */
		schedule_timeout(char_time);
		if (signal_pending(current))
			break;
		if (timeout && time_after(jiffies, orig_jiffies + timeout))
			break;
	}
	current->state = TASK_RUNNING;
#ifdef SERIAL_DEBUG_RS_WAIT_UNTIL_SENT
	//printk("lsr = %d (jiff=%lu)...done\n", lsr, jiffies);
#endif
}

/*
 * rs_hangup() --- called by tty_hangup() when a hangup is signaled.
 */
static void rs_hangup(struct tty_struct *tty)
{
	struct async_struct * info = (struct async_struct *)tty->driver_data;
	struct serial_state *state = info->state;

	printk( "cnxt serial: rs_hangup\n" );

	if (serial_paranoia_check(info, tty->device, "rs_hangup"))
		return;

	state = info->state;

	rs_flush_buffer(tty);
	shutdown(info);
	info->event = 0;
	state->count = 0;
	info->flags &= ~(ASYNC_NORMAL_ACTIVE|ASYNC_CALLOUT_ACTIVE);
	info->tty = 0;
	wake_up_interruptible(&info->open_wait);
}

/*
 * ------------------------------------------------------------
 * rs_open() and friends
 * ------------------------------------------------------------
 */
static int block_til_ready(struct tty_struct *tty, struct file * filp,
			   struct async_struct *info)
{
#if ( LINUX_VERSION_CODE < KERNEL_VERSION( 2, 4, 0 ) )
    struct wait_queue wait = { current, NULL };
#endif

	struct serial_state *state = info->state;
	int		retval;
	int		do_clocal = 0, extra_count = 0;
	unsigned long	flags;

	//printk( "cnxt serial: block_til_ready\n" );

	/*
	 * If the device is in the middle of being closed, then block
	 * until it's done, and then try again.
	 */
	if (tty_hung_up_p(filp) ||
	    (info->flags & ASYNC_CLOSING)) {
		if (info->flags & ASYNC_CLOSING)
		{
			interruptible_sleep_on(&info->close_wait);
			printk( "close_wait UP (1)\n" );
		}
#ifdef SERIAL_DO_RESTART
		return ((info->flags & ASYNC_HUP_NOTIFY) ?
			-EAGAIN : -ERESTARTSYS);
#else
		return -EAGAIN;
#endif
	}

	/*
	 * If this is a callout device, then just make sure the normal
	 * device isn't being used.
	 */
	if (tty->driver.subtype == SERIAL_TYPE_CALLOUT) {
		if (info->flags & ASYNC_NORMAL_ACTIVE)
			return -EBUSY;
		if ((info->flags & ASYNC_CALLOUT_ACTIVE) &&
		    (info->flags & ASYNC_SESSION_LOCKOUT) &&
		    (info->session != current->session))
		    return -EBUSY;
		if ((info->flags & ASYNC_CALLOUT_ACTIVE) &&
		    (info->flags & ASYNC_PGRP_LOCKOUT) &&
		    (info->pgrp != current->pgrp))
		    return -EBUSY;
		info->flags |= ASYNC_CALLOUT_ACTIVE;
		return 0;
	}

	/*
	 * If non-blocking mode is set, or the port is not enabled,
	 * then make the check up front and then exit.
	 */
	if ((filp->f_flags & O_NONBLOCK) ||
	    (tty->flags & (1 << TTY_IO_ERROR))) {
		if (info->flags & ASYNC_CALLOUT_ACTIVE)
			return -EBUSY;
		info->flags |= ASYNC_NORMAL_ACTIVE;
		return 0;
	}

	if (info->flags & ASYNC_CALLOUT_ACTIVE) {
		if (state->normal_termios.c_cflag & CLOCAL)
			do_clocal = 1;
	} else {
		if (tty->termios->c_cflag & CLOCAL)
			do_clocal = 1;
	}

	/*
	 * Block waiting for the carrier detect and the line to become
	 * free (i.e., not in use by the callout).  While we are in
	 * this loop, state->count is dropped by one, so that
	 * rs_close() knows when to free things.  We restore it upon
	 * exit, either normal or abnormal.
	 */
	retval = 0;
#if ( LINUX_VERSION_CODE < KERNEL_VERSION( 2, 4, 0 ) )
	add_wait_queue(&info->open_wait, &wait);
#endif
#ifdef SERIAL_DEBUG_OPEN
	printk("block_til_ready before block: ttys%d, count = %d\n",
		   state->line, state->count);
#endif
	save_flags(flags); cli();
	if (!tty_hung_up_p(filp)) {
		extra_count = 1;
		state->count--;
	}
	restore_flags(flags);
	info->blocked_open++;
	while (1) {
	    /*
		save_flags(flags); cli();
		if (!(info->flags & ASYNC_CALLOUT_ACTIVE) &&
		    (tty->termios->c_cflag & CBAUD))
			serial_out(info, UART_MCR,
				   serial_in(info, UART_MCR) |
				   (UART_MCR_DTR | UART_MCR_RTS));
		restore_flags(flags);
		*/
		current->state = TASK_INTERRUPTIBLE;
		if (tty_hung_up_p(filp) ||
		    !(info->flags & ASYNC_INITIALIZED)) {
#ifdef SERIAL_DO_RESTART
			if (info->flags & ASYNC_HUP_NOTIFY)
				retval = -EAGAIN;
			else
				retval = -ERESTARTSYS;
#else
			retval = -EAGAIN;
#endif
			break;
		}
		if (!(info->flags & ASYNC_CALLOUT_ACTIVE) &&
		    !(info->flags & ASYNC_CLOSING) &&
		    (do_clocal || (serial_in(info, UART_MSR) & UART_MSR_DCD) ))
			break;
		if (signal_pending(current)) {
			retval = -ERESTARTSYS;
			break;
		}
		printk("block_til_ready blocking: count = %d\n", state->count);

#ifdef SERIAL_DEBUG_OPEN
		printk("block_til_ready blocking: ttys%d, count = %d\n",
			   info->line, state->count);
#endif
		schedule();
		printk("block_til_ready waking up\n");
	}
	current->state = TASK_RUNNING;
#if ( LINUX_VERSION_CODE < KERNEL_VERSION( 2, 4, 0 ) )
	remove_wait_queue(&info->open_wait, &wait);
#endif
	if (extra_count)
		state->count++;
	info->blocked_open--;
#ifdef SERIAL_DEBUG_OPEN
	printk("block_til_ready after blocking: ttys%d, count = %d\n",
		   info->line, state->count);
#endif
	if (retval)
		return retval;
	info->flags |= ASYNC_NORMAL_ACTIVE;
	return 0;
}

static int get_async_struct(int line, struct async_struct **ret_info)
{
	struct async_struct *info;
	struct serial_state *sstate;

	//printk( "cnxt serial: get_async_struct\n" );

	sstate = rs_table + line;
	sstate->count++;
	if (sstate->info) {
		*ret_info = sstate->info;
		return 0;
	}
	info = kmalloc(sizeof(struct async_struct), GFP_KERNEL);
	if (!info) {
		sstate->count--;
		return -ENOMEM;
	}
	memset(info, 0, sizeof(struct async_struct));
	info->magic = SERIAL_MAGIC;
	info->port = sstate->port;
	info->flags = sstate->flags;
	info->xmit_fifo_size = sstate->xmit_fifo_size;
	info->line = line;
	info->tqueue.routine = do_softint;
	info->tqueue.data = info;
	info->state = sstate;
	if (sstate->info) {
		kfree(info);
		*ret_info = sstate->info;
		return 0;
	}
	*ret_info = sstate->info = info;
	return 0;
}

/*
 * This routine is called whenever a serial port is opened.  It
 * enables interrupts for a serial port, linking in its async structure into
 * the IRQ chain.   It also performs the serial-specific
 * initialization for the tty structure.
 */
static int rs_open(struct tty_struct *tty, struct file * filp)
{
	struct async_struct	*info;
	int 			retval, line;

	printk( "cnxt serial: rs_open\n" );


	MOD_INC_USE_COUNT;
	line = MINOR(tty->device) - tty->driver.minor_start;
	if ((line < 0) || (line >= NR_PORTS)) {
		MOD_DEC_USE_COUNT;
		return -ENODEV;
	}

	retval = get_async_struct(line, &info);
	if (retval) {
		MOD_DEC_USE_COUNT;
		return retval;
	}
	tty->driver_data = info;
	info->tty = tty;

//	printk("rs_open CRTSCTS is %s\n",(C_CRTSCTS(tty))?"ON":"OFF");
//	printk("rs_open CLOCAL is %s\n",(C_CLOCAL(tty))?"ON":"OFF");
//	printk("rs_open HUPCL is %s\n",(C_HUPCL(tty))?"ON":"OFF");

	if (serial_paranoia_check(info, tty->device, "rs_open")) {
		/* MOD_DEC_USE_COUNT; "info->tty" will cause this */
		printk("serial_paranoia_check(info, tty->device, rs_open) Failed\n");
		return -ENODEV;
	}

#ifdef SERIAL_DEBUG_OPEN
	printk("cnxt: rs_open %s%d, count = %d\n", tty->driver.name, info->line, info->state->count);
#endif

	info->tty->low_latency = (info->flags & ASYNC_LOW_LATENCY) ? 1 : 0;

	if (!tmp_buf)
	{
	    tmp_buf = (unsigned char *)get_free_page(GFP_KERNEL);
	    if (tmp_buf == NULL)
	    {
		/* MOD_DEC_USE_COUNT; "info->tty" will cause this? */
		printk("tmp_buf == NULL\n");
		return -ENOMEM;
	    }
	}


/*	return 0; Shay: ignore everything for now */
	/*
	 * If the port is the middle of closing, bail out now
	 */
	if (tty_hung_up_p(filp) ||
	    (info->flags & ASYNC_CLOSING)) {
		if (info->flags & ASYNC_CLOSING)
		{
			interruptible_sleep_on(&info->close_wait);
			printk( "close_wait UP (2)\n" );
		}
		/* MOD_DEC_USE_COUNT; "info->tty" will cause this? */
#ifdef SERIAL_DO_RESTART
		return ((info->flags & ASYNC_HUP_NOTIFY) ?
			-EAGAIN : -ERESTARTSYS);
#else
		return -EAGAIN;
#endif
	}

/*	return 0; Shay: ignore everything for now */
	/*
	 * Start up serial port
	 */
#ifdef SERIAL_DEBUG_OPEN
	printk("rs_open calling startup...\n");
#endif

	retval = startup(info);
	if (retval) {
		/* MOD_DEC_USE_COUNT; "info->tty" will cause this? */
		return retval;
	}

#ifdef SERIAL_DEBUG_OPEN
	printk("rs_open calling block_til_ready...\n");
#endif

	retval = block_til_ready(tty, filp, info);

	if (retval) {
		/* MOD_DEC_USE_COUNT; "info->tty" will cause this? */
#ifdef SERIAL_DEBUG_OPEN
		printk("rs_open returning after block_til_ready with %d\n", retval);
#endif
		return retval;
	}

	if ((info->state->count == 1) &&
	    (info->flags & ASYNC_SPLIT_TERMIOS)) {
		if (tty->driver.subtype == SERIAL_TYPE_NORMAL)
			*tty->termios = info->state->normal_termios;
		else
			*tty->termios = info->state->callout_termios;
		change_speed(info, 0);
	}

	info->session = current->session;
	info->pgrp = current->pgrp;

#ifdef SERIAL_DEBUG_OPEN
	printk("rs_open ttys%d successful...\n", info->line);
#endif
	return 0;
}

/*
 * /proc fs routines....
 */

static inline int line_info(char *buf, struct serial_state *state)
{
	struct async_struct *info = state->info, scr_info;
	char	stat_buf[30], control, status;
	int	ret;
	unsigned long flags;

	//printk( "cnxt serial: line_info\n" );

	ret = sprintf(buf, "%d: uart:%s port:%x irq:%d",
		      state->line, uart_config[state->type].name,
		      state->port, state->irq);

	if (!state->port || (state->type == PORT_UNKNOWN)) {
		ret += sprintf(buf+ret, "\n");
		return ret;
	}

	/*
	 * Figure out the current RS-232 lines
	 */
	if (!info) {
		info = &scr_info;	/* This is just for serial_{in,out} */

		info->magic = SERIAL_MAGIC;
		info->port = state->port;
		info->flags = state->flags;
		info->quot = 0;
		info->tty = 0;
	}
	save_flags(flags); cli();
	status = serial_in(info, UART_MSR);
	control = info ? info->MCR : serial_in(info, UART_MCR);
	restore_flags(flags);

	stat_buf[0] = 0;
	stat_buf[1] = 0;
	if (control & UART_MCR_RTS)
		strcat(stat_buf, "|RTS");
	if (status & UART_MSR_CTS)
		strcat(stat_buf, "|CTS");
	if (control & UART_MCR_DTR)
		strcat(stat_buf, "|DTR");
	if (status & UART_MSR_DSR)
		strcat(stat_buf, "|DSR");
	if (status & UART_MSR_DCD)
		strcat(stat_buf, "|CD");
	if (status & UART_MSR_RI)
		strcat(stat_buf, "|RI");

	if (info->quot) {
		ret += sprintf(buf+ret, " baud:%d",
			       state->baud_base / info->quot);
	}

	ret += sprintf(buf+ret, " tx:%d rx:%d",
		      state->icount.tx, state->icount.rx);

	if (state->icount.frame)
		ret += sprintf(buf+ret, " fe:%d", state->icount.frame);

	if (state->icount.parity)
		ret += sprintf(buf+ret, " pe:%d", state->icount.parity);

	if (state->icount.brk)
		ret += sprintf(buf+ret, " brk:%d", state->icount.brk);

	if (state->icount.overrun)
		ret += sprintf(buf+ret, " oe:%d", state->icount.overrun);

	/*
	 * Last thing is the RS-232 status lines
	 */
	ret += sprintf(buf+ret, " %s\n", stat_buf+1);
	return ret;
}

int rs_read_proc(char *page, char **start, off_t off, int count,
		 int *eof, void *data)
{
	int len = 0, l;
	off_t	begin = 0;

	//printk( "cnxt serial: rs_read_proc\n" );

	len += sprintf(page, "serinfo:1.0 driver:%s\n", serial_version);

	l = line_info(page + len, rs_table /* &rs_table[0] */);
	len += l;
	if (len+begin > off+count)
		goto done;
	if (len+begin < off) {
		begin += len;
		len = 0;
	}

	*eof = 1;
done:
	if (off >= len+begin)
		return 0;
	*start = page + (begin-off);
	return ((count < begin+len-off) ? count : begin+len-off);
}

/*
 * ---------------------------------------------------------------------
 * rs_init() and friends
 *
 * rs_init() is called at boot-time to initialize the serial driver.
 * ---------------------------------------------------------------------
 */

/*
 * This routine prints out the appropriate serial driver version
 * number, and identifies which options were configured into this
 * driver.
 */
static _INLINE_ void show_serial_version(void)
{
 	//printk(KERN_INFO "%s version %s with", serial_name, serial_version);
#ifdef CONFIG_HUB6
	//printk(" HUB-6");
#define SERIAL_OPT
#endif
#ifdef CONFIG_SERIAL_MANY_PORTS
	//printk(" MANY_PORTS");
#define SERIAL_OPT
#endif
#ifdef CONFIG_SERIAL_MULTIPORT
	//printk(" MULTIPORT");
#define SERIAL_OPT
#endif
#ifdef CONFIG_SERIAL_SHARE_IRQ
	//printk(" SHARE_IRQ");
#define SERIAL_OPT
#endif
#ifdef CONFIG_SERIAL_DETECT_IRQ
	//printk(" DETECT_IRQ");
#define SERIAL_OPT
#endif
#ifdef SERIAL_OPT
	//printk(" enabled\n");
#else
	//printk(" no serial options enabled\n");
#endif
#undef SERIAL_OPT
}

/*
 * This routine is called by rs_init() to initialize a specific serial
 * port.  It determines what type of UART chip this serial port is
 * using: 8250, 16450, 16550, 16550A.  The important question is
 * whether or not this UART is a 16550A, since this will determine
 * whether or not we can use its FIFO features.
 */
static void autoconfig(struct serial_state * state)
{
	struct async_struct *info, scr_info;
/*	unsigned long flags; */

	//printk( "cnxt serial: autoconfig\n" );

	state->type = PORT_UNKNOWN;

	if (!state->port)
		return;

	info = &scr_info;	/* This is just for serial_{in,out} */

	info->magic = SERIAL_MAGIC;
	info->port = state->port;
	info->flags = state->flags;

/*	save_flags(flags); cli(); */ /* ??? */

	/* Shay: skipped existence tests */

	state->type = PORT_16550A; /* Shay: use 16550A so we can use it's fifo features */

	state->xmit_fifo_size =	uart_config[state->type].dfl_xmit_fifo_size;

/*	request_region(info->port,8,"serial(auto)"); */ /* Shay: registering the hardware, should be done by the hardware driver */

/*	restore_flags(flags); */ /* ??? */
}

int register_cnxt_serial(struct serial_struct *req);
void unregister_cnxt_serial(int line);

EXPORT_SYMBOL(register_cnxt_serial);
EXPORT_SYMBOL(unregister_cnxt_serial);

/*
 * the function registers and unregisters a dummy char device to get a free major number.
 */
int get_major_number( void )
{
	int result;

	result = register_chrdev( 0, "noname", 0 );
	if ( result <= 0 )
		return 0;

	unregister_chrdev( result, "noname" );

	return result;
}

/*
 * The serial driver boot-time initialization code!
 */
#if ( LINUX_VERSION_CODE < KERNEL_VERSION( 2, 4, 0 ) )
__initfunc(int rs_init(void))
#else
static int __init rs_init(void)
#endif
{
	int i;
	struct serial_state * state;
/*	extern void atomwide_serial_init (void); */
/*	extern void dualsp_serial_init (void); */

/*#ifdef CONFIG_ATOMWIDE_SERIAL
	atomwide_serial_init ();
#endif
#ifdef CONFIG_DUALSP_SERIAL
	dualsp_serial_init ();
#endif*/

	//printk( "cnxt serial: rs_init\n" );

	init_bh(SERIAL_BH, do_serial_bh);
	init_timer(&serial_timer);
	serial_timer.function = rs_timer;
	mod_timer(&serial_timer, jiffies + RS_STROBE_TIME);

	for (i = 0; i < NR_IRQS; i++) {
		IRQ_ports[i] = 0;
		IRQ_timeout[i] = 1;
	}

	show_serial_version();

	/* Initialize the tty_driver structure */

	memset(&serial_driver, 0, sizeof(struct tty_driver));
	serial_driver.magic = TTY_DRIVER_MAGIC;
	serial_driver.driver_name = "serial";
	serial_driver.name = "cnxt*";			  /* Shay: for the good order out modems should be called cnxt0 etc .. */
	serial_driver.major = get_major_number(); /* Shay: request dynamic assignment of major number */
	if ( serial_driver.major == 0 )
		panic( "Serial: Couldn't assign major number" );
	serial_driver.minor_start = 64;
	serial_driver.num = 1;					  /* Shay: support for one device */
	serial_driver.type = TTY_DRIVER_TYPE_SERIAL;
	serial_driver.subtype = SERIAL_TYPE_NORMAL;
	serial_driver.init_termios = tty_std_termios;
	serial_driver.init_termios.c_cflag =
		B9600 | CS8 | CREAD | HUPCL | CLOCAL;
	serial_driver.flags = TTY_DRIVER_REAL_RAW;
	serial_driver.refcount = &serial_refcount;
	serial_driver.table = serial_table;
	serial_driver.termios = serial_termios;
	serial_driver.termios_locked = serial_termios_locked;

	serial_driver.open = rs_open;
	serial_driver.close = rs_close;
	serial_driver.write = rs_write;
	serial_driver.put_char = rs_put_char;
	serial_driver.flush_chars = rs_flush_chars;
	serial_driver.write_room = rs_write_room;
	serial_driver.chars_in_buffer = rs_chars_in_buffer;
	serial_driver.flush_buffer = rs_flush_buffer;
	serial_driver.ioctl = rs_ioctl;
	serial_driver.throttle = rs_throttle;
	serial_driver.unthrottle = rs_unthrottle;
	serial_driver.send_xchar = rs_send_xchar;
	serial_driver.set_termios = rs_set_termios;
	serial_driver.stop = rs_stop;
	serial_driver.start = rs_start;
	serial_driver.hangup = rs_hangup;
	serial_driver.break_ctl = rs_break;
	serial_driver.wait_until_sent = rs_wait_until_sent;
	serial_driver.read_proc = rs_read_proc;

	/*
	 * The callout device is just like normal device except for
	 * major number and the subtype code.
	 */
	callout_driver = serial_driver;
	callout_driver.name = "cua";
	callout_driver.major = TTYAUX_MAJOR;
	callout_driver.subtype = SERIAL_TYPE_CALLOUT;
	callout_driver.read_proc = 0;
	callout_driver.proc_entry = 0;

	if (tty_register_driver(&serial_driver))
		panic("Couldn't register serial driver\n");
	if (tty_register_driver(&callout_driver))
		panic("Couldn't register callout driver\n");

	/* Shay: removed multi port support, only config one port */
	state = rs_table;	/* state = &rs_table[0] */
	state->magic = SSTATE_MAGIC;
	state->line = 0;
	state->type = PORT_UNKNOWN;
	state->custom_divisor = 0;
	state->close_delay = 5*HZ/10;
	state->closing_wait = 30*HZ;
	state->callout_termios = callout_driver.init_termios;
	state->normal_termios = serial_driver.init_termios;
	state->icount.cts = state->icount.dsr =
		state->icount.rng = state->icount.dcd = 0;
	state->icount.rx = state->icount.tx = 0;
	state->icount.frame = state->icount.parity = 0;
	state->icount.overrun = state->icount.brk = 0;
	state->irq = irq_cannonicalize(state->irq);
	if (!check_region(state->port,8)) {
		if (state->flags & ASYNC_BOOT_AUTOCONF)
			autoconfig(state);
	}

	state->irq = 0; /* Shay: use irq 0 for now ?!? */

	if ( init_com_if() == 0 )
		printk( "imale .. (can't open desired interface)\n" );

	return 0;
}

/*
 * register_cnxt_serial and unregister_cnxt_serial allows for serial ports to be
 * configured at run-time, to support PCMCIA modems.
 */
int register_cnxt_serial(struct serial_struct *req)
{
	int i;
	unsigned long flags;
	struct serial_state *state;

	//printk( "cnxt serial: register_cnxt_serial\n" );

	save_flags(flags);
	cli();
	for (i = 0; i < NR_PORTS; i++) {
		if (rs_table[i].port == req->port)
			break;
	}
	if (i == NR_PORTS) {
		for (i = 0; i < NR_PORTS; i++)
			if ((rs_table[i].type == PORT_UNKNOWN) &&
			    (rs_table[i].count == 0))
				break;
	}
	if (i == NR_PORTS) {
		restore_flags(flags);
		return -1;
	}
	state = &rs_table[i];
	if (rs_table[i].count) {
		restore_flags(flags);
		//printk("Couldn't configure serial #%d (port=%d,irq=%d): "
		    //   "device already open\n", i, req->port, req->irq);
		return -1;
	}
	state->irq = req->irq;
	state->port = req->port;
	state->flags = req->flags;

	autoconfig(state);
	if (state->type == PORT_UNKNOWN) {
		restore_flags(flags);
		//printk("register_cnxt_serial(): autoconfig failed\n");
		return -1;
	}
	restore_flags(flags);

	if ((state->flags & ASYNC_AUTO_IRQ) && (state->port != 0))
		state->irq = 0; /* Shay: use 0 for now ?!? */

	//printk(KERN_INFO "tty%02d at 0x%04x (irq = %d) is a %s\n",
	 //      state->line, state->port, state->irq,
	   //    uart_config[state->type].name);
	return state->line;
}

void unregister_cnxt_serial(int line)
{
	unsigned long flags;
	struct serial_state *state = &rs_table[line];

	//printk( "cnxt serial: unregister_cnxt_serial\n" );

	save_flags(flags);
	cli();
	if (state->info && state->info->tty)
		tty_hangup(state->info->tty);
	state->type = PORT_UNKNOWN;
	//printk(KERN_INFO "tty%02d unloaded\n", state->line);
	restore_flags(flags);
}

#ifdef MODULE
int init_module(void)
{
    printk("serial init_module\n");
	printk("  compiled on %s %s\n", __DATE__, __TIME__ );
    return rs_init();
}

void cleanup_module(void)
{
	unsigned long flags;
	int e1, e2;

	printk("Unloading %s: version %s\n", serial_name, serial_version);
	
	del_timer(&serial_timer);
	save_flags(flags);
	cli();
    remove_bh(SERIAL_BH);
	if ((e1 = tty_unregister_driver(&serial_driver)))
	{
	    printk("SERIAL: failed to unregister serial driver (%d)\n",e1);
	}
	if ((e2 = tty_unregister_driver(&callout_driver)))
	{
	    printk("SERIAL: failed to unregister callout driver (%d)\n",e2);
	}

#ifndef SKELETON_SERIAL
	if ( NULL != ExtendedInfo.pComCtrl )
	{
		ExtendedInfo.pComCtrl->Close(ExtendedInfo.hComCtrl);
		ExtendedInfo.pComCtrl->Destroy(ExtendedInfo.hComCtrl);
	}
	ExtendedInfo.pComCtrl = NULL;
	ExtendedInfo.hComCtrl = NULL;
#endif

	restore_flags(flags);

/*	release_region(rs_table -> port, 8); */ /* Shay: should be done by the hardware driver */
	if (tmp_buf)
	{
	    free_page((unsigned long) tmp_buf);
	    tmp_buf = NULL;
	}
}
#endif /* MODULE */

#ifndef SKELETON_SERIAL
static void rs_eventNotify(PVOID hport, UINT32 dwEvtMask)
{
//	printk( "rs_evt: %x\n", dwEvtMask );
	static UINT32 old_uart_msr, uart_msr;

	old_uart_msr = uart_msr = serial_in( NULL, UART_MSR );

#if 0
    if(dwEvtMask & COMCTRL_EVT_RXCHAR)
    {
        dwEvtMask &= ~COMCTRL_EVT_RXCHAR;
        ReadFromMC((HANDLE)hport);
    }

    if( dwEvtMask & COMCTRL_EVT_TXCHAR ||
        dwEvtMask & COMCTRL_EVT_TXEMPTY   )
    {
        dwEvtMask &= ~(COMCTRL_EVT_TXCHAR | COMCTRL_EVT_TXEMPTY);
        WriteToMC((HANDLE)hport);
    }
#endif

    if(dwEvtMask & COMCTRL_EVT_CTS)
    {
/*  		printk( "rs_evt: CTS %s\n", */
/*  				(dwEvtMask & COMCTRL_EVT_CTSS) ? "on" : "off" ); */
        if ( dwEvtMask & COMCTRL_EVT_CTSS )
            uart_msr |= UART_MSR_CTS; /* UART_MSR_CTS --> CTS ON */
        else
            uart_msr &= ~UART_MSR_CTS;
    }

    if(dwEvtMask & COMCTRL_EVT_DSR)
    {
/*  		printk( "rs_evt: DSR %s\n", */
/*  				(dwEvtMask & COMCTRL_EVT_DSRS) ? "on" : "off" ); */
        if(dwEvtMask & COMCTRL_EVT_DSRS)
            uart_msr |= UART_MSR_DSR; /* UART_MSR_DSR --> DSR ON */
        else
            uart_msr &= ~UART_MSR_DSR;
    }

    if(dwEvtMask & COMCTRL_EVT_RLSD)
    {
/*    		printk( "rs_evt: DCD %s\n", */
/*    				(dwEvtMask & COMCTRL_EVT_RLSDS) ? "on" : "off" ); */
        if(dwEvtMask & COMCTRL_EVT_RLSDS)
            uart_msr |= UART_MSR_DCD; /* UART_MSR_DCD --> RLSD ON */
        else
            uart_msr &= ~UART_MSR_DCD;
		uart_msr |= UART_MSR_DDCD;
    }

	if ( ( (uart_msr & UART_MSR_RI) && !(dwEvtMask & COMCTRL_EVT_RING) ) ||
		 ( !(uart_msr & UART_MSR_RI) && (dwEvtMask & COMCTRL_EVT_RING) ) )
	{
/*  		printk( "rs_evt: RING %s\n", */
/*  				(dwEvtMask & COMCTRL_EVT_RING) ? "on" : "off" ); */
		if(dwEvtMask & COMCTRL_EVT_RING)
			uart_msr |= UART_MSR_RI; /* UART_MSR_RI --> Ring Indicator */
		else // if(EV_RingTe)
			uart_msr &= ~UART_MSR_RI;
	}

	if ( old_uart_msr != uart_msr )
	{
/*  		printk( "--> Updating UART_MSR to 0x%x\n", (unsigned int) uart_msr ); */

		/* DANGER!!! serial_out should get async_struct *info, not NULL!!!
		   But we don't know it here. Actually, we aren't multi-port/irq */
		serial_out( NULL, UART_MSR, uart_msr );
	}

}  /* rs_eventNotify */
#endif

/* Shay: below is stuff that was ripped from octopus' PortDrv */
/*  get the comm controler interface.
	the function is called before bothering the kernel so if we fail
	there's no need to release/free/do anything (except the critical session *sigh*)
*/
int init_com_if( void )
{

#ifndef SKELETON_SERIAL

    PORT_EVENT_HANDLER EvtHandler;

    ExtendedInfo.pComCtrl = ComCtrlGetInterface();
    if(NULL == ExtendedInfo.pComCtrl)
    {
        return 0;
    }

    if(COMCTRL_VERSION != ExtendedInfo.pComCtrl->GetInterfaceVersion())
    {
        return 0;
    }

    ExtendedInfo.hComCtrl = ExtendedInfo.pComCtrl->Create();

    if(NULL == ExtendedInfo.hComCtrl)
    {
        return 0;
    }

    EvtHandler.pfnCallback = rs_eventNotify;
    EvtHandler.pRef        = ExtendedInfo.pComCtrl;

    if ( ( COM_STATUS_SUCCESS !=
		   ExtendedInfo.pComCtrl->Configure( ExtendedInfo.hComCtrl,
											 COMCTRL_CONFIG_EVENT_HANDLER,
											 (PVOID) &EvtHandler ) ) ||
		 ( COM_STATUS_SUCCESS !=
		   ExtendedInfo.pComCtrl->Init(ExtendedInfo.hComCtrl) ) )
    {
        ExtendedInfo.pComCtrl->Destroy(ExtendedInfo.hComCtrl);
        return 0;
    }

#endif

    return 1; /* success :) */
}

void cnxt_task( void *data )
{
    rs_interrupt_single(0, NULL, NULL);
}

